import dataclasses
import os
from typing import Optional, cast

import pytest
import pytorch_lightning as pl
import torch
from rdkit import Chem
from rdkit.Chem.Descriptors import NumAromaticRings
from syntheseus.interface.molecule import Molecule
from torch import Tensor, nn
from torch.utils.data import DataLoader, Dataset
from torch_geometric.data import Batch

from retrochimera.encoders.chem import ChemEncoder
from retrochimera.encoders.configs import GNNEncoderConfig
from retrochimera.encoders.gnn import GNNEncoder


def assert_shape(t: Optional[Tensor], expected_shape: list[int]) -> None:
    assert t is not None
    assert isinstance(t, Tensor)
    assert list(t.shape) == expected_shape


class Model(pl.LightningModule):
    def __init__(self, encoder) -> None:
        super().__init__()

        self.encoder = encoder
        self.proj = nn.Linear(encoder.mol_out_channels, 1)

    def forward(self, batch: Batch):
        return self.proj(self.encoder(batch).mol_outputs).squeeze(-1)

    def ttv_step(self, batch, step_name):
        inputs, targets = batch
        outputs = self.forward(inputs)
        loss = torch.nn.functional.mse_loss(outputs, targets)

        self.log(f"{step_name}_loss", loss)
        return loss

    def training_step(self, batch, batch_idx):
        return self.ttv_step(batch, "train")

    def validation_step(self, batch, batch_idx):
        return self.ttv_step(batch, "val")

    def test_step(self, batch, batch_idx):
        return self.ttv_step(batch, "test")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer


@pytest.fixture
def train_mols() -> list[Molecule]:
    return [
        Molecule(smiles.rstrip())
        for smiles in open(
            os.path.join(os.path.dirname(__file__), "..", "testdata", "guacamol_100_smiles.txt")
        )
    ]


def get_default_chem_encoder() -> ChemEncoder:
    return ChemEncoder(fingerprint_dim=2048)


def get_default_gnn_config() -> GNNEncoderConfig:
    config = GNNEncoderConfig(
        atom_out_channels=16, bond_out_channels=24, mol_out_channels=32, aggregation_dropout=0.0
    )
    config.gnn_kwargs["hidden_channels"] = 8
    config.gnn_kwargs["num_layers"] = 1

    return config


def get_default_gnn_encoder() -> GNNEncoder:
    return GNNEncoder(**dataclasses.asdict(get_default_gnn_config()))


def get_graphium_gnn_encoder() -> GNNEncoder:
    config = get_default_gnn_config()
    config.gnn_class = "GPS_PNA"
    config.bond_out_channels = None

    return GNNEncoder(**dataclasses.asdict(config))


@pytest.mark.parametrize(
    "encoder_fn", [get_default_chem_encoder, get_default_gnn_encoder, get_graphium_gnn_encoder]
)
class TestEncoders:
    def test_encode_mols(self, encoder_fn) -> None:
        mols = [Molecule(smiles) for smiles in ["c1ccccc1N=O", "CCC"]]

        encoder = encoder_fn()
        output = encoder.forward_raw(mols)

        rdkit_mols = [mol.rdkit_mol for mol in mols]
        num_atoms = sum(mol.GetNumAtoms() for mol in rdkit_mols)
        num_bonds = sum(mol.GetNumBonds() for mol in rdkit_mols)
        num_mols = len(mols)

        # How many outputs we expect for each "kind" of output.
        num_outputs_per_level = {"atom": num_atoms, "bond": 2 * num_bonds, "mol": num_mols}

        for level, expected_num_outputs in num_outputs_per_level.items():
            num_channels = getattr(encoder, f"{level}_out_channels")

            if num_channels is not None:
                output_for_level = getattr(output, f"{level}_outputs")
                assert_shape(output_for_level, [expected_num_outputs, num_channels])

    def test_train(self, encoder_fn, train_mols: list[Chem.Mol]):
        for _ in range(20):
            encoder = encoder_fn()
            train_mols_processed = [encoder.preprocess(mol) for mol in train_mols]

            # Set the targets to the number of aromatic rings, which should be learnable by any encoder.
            train_targets = torch.as_tensor(
                [NumAromaticRings(mol.rdkit_mol) for mol in train_mols],
                dtype=torch.float32,
            )

            # A list implements the `Dataset` interface via duck typing.
            dataset = cast(Dataset, list(zip(train_mols_processed, train_targets)))

            # Compute the loss for a baseline that always predicts the mean of all the targets.
            predict_the_mean_loss = torch.nn.functional.mse_loss(
                train_targets, train_targets.mean()
            )

            def collate_fn(samples: list[tuple]):
                inputs = [input for input, _ in samples]
                targets = [target for _, target in samples]
                return encoder.collate(inputs), torch.as_tensor(targets)

            model = Model(encoder)
            data_loader = DataLoader(dataset, batch_size=8, shuffle=True, collate_fn=collate_fn)

            trainer = pl.Trainer(max_epochs=20, logger=False, enable_checkpointing=False)

            # Before training the predictions should be no better than a naive baseline unless we
            # got lucky with initialization. In the latter case we stay in the loop and retry.
            [results_before] = trainer.test(model, dataloaders=data_loader)
            if results_before["test_loss"] > predict_the_mean_loss:
                break

        trainer.fit(model, train_dataloaders=data_loader)
        model.eval()

        # After training the predictions should be better than the baseline.
        [results_after] = trainer.test(model, dataloaders=data_loader)
        assert results_after["test_loss"] < 0.75 * predict_the_mean_loss

    def test_batching(self, encoder_fn) -> None:
        encoder = encoder_fn()
        mols = [Molecule(smiles) for smiles in ["CN", "CC"]]

        encoder.eval()
        outputs_unbatched = [encoder.forward_raw([mol]) for mol in mols]
        outputs_batched = encoder.forward_raw(mols)

        for level in ["atom", "bond", "mol"]:
            if getattr(encoder, f"{level}_out_channels") is None:
                continue

            encodings_unbatched = [
                getattr(output, f"{level}_outputs") for output in outputs_unbatched
            ]
            encodings_batched = getattr(outputs_batched, f"{level}_outputs")

            # Encodings should depend on the molecule.
            assert not torch.allclose(encodings_unbatched[0], encodings_unbatched[1], atol=1e-4)

            # Encodings should not depend on whether there were other molecules in the batch.
            assert torch.allclose(torch.cat(encodings_unbatched), encodings_batched, atol=1e-4)



from abc import abstractproperty
from pathlib import Path
from typing import Iterable, Optional, TypeVar, Union

from syntheseus.reaction_prediction.data import dataset as dataset_syntheseus
from syntheseus.reaction_prediction.data.dataset import DataFold
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample

from retrochimera.chem.rules import RuleBase

SampleType = TypeVar("SampleType", bound=ReactionSample)


class ReactionDataset(dataset_syntheseus.ReactionDataset[SampleType]):
    """Dataset holding raw reactions split into folds (together with a rulebase, if applicable).

    Depending on implementation, the reactions can either be held on disk or in memory.
    """

    @abstractproperty
    def rulebase(self) -> RuleBase:
        pass


class DiskReactionDataset(dataset_syntheseus.DiskReactionDataset[SampleType], ReactionDataset):
    def __init__(
        self,
        data_dir: Union[str, Path],
        sample_cls: type[SampleType],
        num_processes: int = 0,
        rulebase_min_rule_support: Optional[int] = None,
        rulebase_max_num_rules: Optional[int] = None,
    ):
        super().__init__(data_dir, sample_cls=sample_cls, num_processes=num_processes)

        self._rulebase: Optional[RuleBase] = None
        self._rulebase_min_rule_support = rulebase_min_rule_support
        self._rulebase_max_num_rules = rulebase_max_num_rules

    @property
    def rulebase(self) -> RuleBase:
        if self._rulebase is None:
            self._rulebase = RuleBase.load_from_file(
                dir=self._data_dir,
                min_rule_support=self._rulebase_min_rule_support,
                max_num_rules=self._rulebase_max_num_rules,
            )

        return self._rulebase


class InMemoryReactionDataset(ReactionDataset[SampleType]):
    def __init__(
        self, samples: dict[DataFold, list[SampleType]], rulebase: Optional[RuleBase] = None
    ):
        self._samples = samples
        self._rulebase = rulebase
        self._num_samples = {fold: len(samples_list) for fold, samples_list in samples.items()}

    def __getitem__(self, fold: DataFold) -> Iterable[SampleType]:
        return self._samples.get(fold, [])

    def get_num_samples(self, fold: DataFold) -> int:
        return self._num_samples.get(fold, 0)

    @property
    def rulebase(self) -> RuleBase:
        if self._rulebase is None:
            raise ValueError("Rulebase was not provided")

        return self._rulebase



from pathlib import Path
from typing import Union

import pandas as pd

from retrochimera.data.dataset import DataFold


def load_raw_reactions_file(path: Union[str, Path]) -> list[str]:
    if str(path).endswith(".csv"):
        # USPTO-style format.
        data = pd.read_csv(path)
        column_name = "reactants>reagents>production"

        if column_name not in data.columns:
            raise ValueError(f"Column {column_name} not found in {path}")

        return data[column_name].values
    elif str(path).endswith(".smi"):
        # Generic format with one reaction SMILES per line, with optionally extra into after \t.
        with open(path) as f:
            return [line.rstrip() for line in f]
    else:
        raise ValueError(f"Unrecognized file extension in {path}")


def load_raw_reactions_files(dir: Union[str, Path]) -> dict[DataFold, list[str]]:
    """Load reactions for each fold from files in the given directory.

    Args:
        dir: Directory containing the reaction files.

    Returns:
        Dictionary mapping folds to lists of reactions.
    """
    fold_to_path: dict[DataFold, Path] = {}
    for fold in DataFold:
        matching_paths = sum(
            [list(Path(dir).glob(f"*{fold.value}*.{ext}")) for ext in ["csv", "smi"]], []
        )

        if not matching_paths:
            raise ValueError(f"No files found for fold {fold.value}")

        if len(matching_paths) > 1:
            raise ValueError(
                f"Multiple files found for fold {fold.value}: {[str(f) for f in matching_paths]}"
            )

        fold_to_path[fold] = matching_paths[0]

    if len(set(path.suffix for path in fold_to_path.values())) > 1:
        raise ValueError("Files for different folds have inconsistent formats")

    return {fold: load_raw_reactions_file(path) for fold, path in fold_to_path.items()}



import pytest
from rdkit import Chem
from syntheseus.interface.molecule import Molecule

from retrochimera.chem.rewrite import Rewrite


@pytest.mark.parametrize(
    "template, num_mapped",
    [
        ("[c:10]-[N&H2&+0&D1:2]>>[c:10]-[N&H0&+&D3:2](=O)-[O&-:5]", 2),
        ("[c:0]-[N&H2&+0&D1:2]>>[c:0]-[N&H1&+0&D2:2]-C(=O)-O-C(-C)(-C)-C", 2),
        ("[C:1]-[C&H2&+0&D2:33]-[O&H1&+0&D1:20]>>[C:1]-[C&H0&+0&D3:33](=[O&H0&+0&D1:20])-O-C-C", 3),
        ("[C:1]-[C&H1&+0&D2:3]=[O&H0&+0&D1:2]>>[C:1]-[C&H1&+0&D3:3]1-[O&H0&+0&D2:2]-C-C-O-1", 3),
    ],
)
def test_rewrite(template, num_mapped) -> None:
    rewrite = Rewrite.from_rdkit(template)

    def get_atoms_list(smarts: str):
        return list(Chem.MolFromSmarts(smarts).GetAtoms())

    atoms_lhs = get_atoms_list(rewrite.lhs)
    atoms_rhs = get_atoms_list(rewrite.rhs)

    assert len(rewrite.mapping) == num_mapped

    # Sanity check: atom mapping preserves the atom type.
    for idx_lhs, idx_rhs in rewrite.mapping:
        assert atoms_lhs[idx_lhs].GetAtomicNum() == atoms_rhs[idx_rhs].GetAtomicNum()

    # Matches all of the tested templates exactly once.
    input_mol = Molecule("Nc1ccc(CCO)c(CC=O)c1")
    results = rewrite.apply(input_mol)

    assert len(results) == 1
    [result] = results

    assert len(result.metadata["localizations"]) == 1

    [localization] = result.metadata["localizations"]
    assert len(localization) == num_mapped

    # Sanity check: localization is consistent with atom type.
    atoms_input = list(input_mol.rdkit_mol.GetAtoms())
    for idx_lhs, idx_input in enumerate(localization):
        assert atoms_lhs[idx_lhs].GetAtomicNum() == atoms_input[idx_input].GetAtomicNum()

    # Matches the templates exactly twice, but with the same result.
    input_mol = Molecule("Nc1ccc(CCO)c(CC=O)c1.Nc1ccc(CCO)c(CC=O)c1")
    results = rewrite.apply(input_mol)

    assert len(results) == 1

    for result in results:
        assert len(result.metadata["localizations"]) == 2

    # Matches the templates exactly twice, with different result.
    input_mol = Molecule("Nc1ccc(CCO)c(CC=O)c1.Nc1c(Cl)cc(CCO)c(CC=O)c1")
    results = rewrite.apply(input_mol)

    assert len(results) == 2

    for result in results:
        assert len(result.metadata["localizations"]) == 1

    # Run on a molecule that shouldn't match any of the tested tempaltes.
    assert not rewrite.apply(Molecule("NC=O"))



from collections import defaultdict
from dataclasses import dataclass

from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from syntheseus.reaction_prediction.utils.misc import cpu_count

from retrochimera.data.processing import parse_samples
from retrochimera.data.utils import load_raw_reactions_file
from retrochimera.utils.logging import get_logger
from retrochimera.utils.misc import silence_rdkit_warnings

logger = get_logger(__name__)


@dataclass
class DeduplicateDatasetConfig:
    """Config for deduplicating a given dataset."""

    input_path: str  # Path to raw dataset, either Pistachio format (.smi) or USPTO format (.csv).
    output_path: str = "pistachio_deduplicated.smi"  # Path to a *.smi file for saving the output.
    num_processes: int = cpu_count()  # Number of processes to use for processing the SMILES.


def deduplicate_dataset(config: DeduplicateDatasetConfig) -> None:
    path = config.input_path
    raw_data = load_raw_reactions_file(path)
    unique_samples: dict[str, set[str]] = defaultdict(set)

    logger.info(f"Processing {len(raw_data)} samples")
    data = parse_samples(raw_data, num_processes=config.num_processes)
    logger.info(f"Parsed {len(data)} valid samples\n")

    for d in data:
        unique_samples[d.reaction_smiles].add(d.metadata["original_str"])  # type: ignore
    logger.info(f"Number of unique reaction SMILES: {len(unique_samples)}")

    # Save unique reaction SMILES.
    with open(config.output_path, "wt") as f:
        for original_str in unique_samples.values():
            f.write(f"{original_str.pop()}\n")


def main() -> None:
    silence_rdkit_warnings()

    config: DeduplicateDatasetConfig = cli_get_config(
        argv=None, config_cls=DeduplicateDatasetConfig
    )

    logger.info(f"Removing duplicates with the following config: {config}")

    deduplicate_dataset(config)


if __name__ == "__main__":
    main()



import dataclasses
import tempfile
from pathlib import Path
from typing import Any, Union

import pytest
from syntheseus.cli.eval_single_step import get_results

from retrochimera.cli.train import (
    ModelClass,
    TemplateClassificationChemConfig,
    TemplateClassificationGNNConfig,
    TrainConfig,
    train,
)
from retrochimera.data.dataset import DataFold, ReactionDataset
from retrochimera.encoders.configs import GNNEncoderConfig
from retrochimera.inference.template_classification import TemplateClassificationModel
from retrochimera.models.template_classification import MCCModel
from retrochimera.utils.training import preprocess_and_save
from tests.chem.test_rules import check_results_against_targets


@pytest.mark.parametrize(
    "encoder_class,encoder_kwargs",
    [
        ("ChemEncoder", {"fingerprint_dim": 1024}),
        ("GNNEncoder", dataclasses.asdict(GNNEncoderConfig()) | {"aggregation_dropout": 0.0}),
    ],
)
def test_workflow_end_to_end(
    encoder_class: str, encoder_kwargs: dict[str, Any], reaction_dataset: ReactionDataset
) -> None:
    """Test the full workflow of training an `MCCModel`, loading it, and running inference."""

    with tempfile.TemporaryDirectory() as tmp_dir:
        config = TrainConfig(
            processed_data_dir=tmp_dir,
            model_class=ModelClass[f"TemplateClassification{encoder_class[:-7]}"],
            checkpoint_dir=tmp_dir,
        )

        model_config: Union[TemplateClassificationChemConfig, TemplateClassificationGNNConfig]
        if encoder_class == "ChemEncoder":
            model_config = config.template_classification_chem_config
        elif encoder_class == "GNNEncoder":
            model_config = config.template_classification_gnn_config
        else:
            raise ValueError(f"Encoder class {encoder_class} not recognized")

        model_config.mlp.dropout = 0.0
        model_config.mlp.hidden_dim = 128
        model_config.training.n_epochs = 30
        model_config.training.accelerator = "cpu"

        model = MCCModel(
            encoder_class=encoder_class,
            encoder_kwargs=encoder_kwargs,
            hidden_dim=model_config.mlp.hidden_dim,
            n_classes=len(reaction_dataset.rulebase),
            n_hidden_layers=model_config.mlp.n_layers,
        )
        reaction_dataset.rulebase.save_to_file(dir=config.checkpoint_dir)
        model.set_rulebase(rulebase=reaction_dataset.rulebase, rulebase_dir=config.checkpoint_dir)

        # Run data preprocessing and training.
        processed_data_path = str(Path(tmp_dir) / "data.h5")

        preprocess_and_save(
            save_path=processed_data_path, dataset=reaction_dataset, model=model, num_processes=0
        )
        results = train(
            model=model, data_path=processed_data_path, config=config, model_config=model_config
        )

        # Check that the training has worked.
        assert len(results.test_results) > 0
        assert results.test_results[0]["test_acc"] > 0.75

        # Build the wrapper, and exit the context (the temporary dir will not be needed unless the
        # rule application processes need to be restarted which shouldn't happen here).
        model_wrapper = TemplateClassificationModel(model_dir=tmp_dir, device="cpu")

    train_reactant_smiles = []
    train_product_mols = []

    for datapoint in reaction_dataset[DataFold.TRAIN]:
        train_reactant_smiles.append(datapoint.reactants_str)
        train_product_mols.append(list(datapoint.products)[0])

    # Run inference using a lower-level API specific to `TemplateClassificationModel`.
    inf_results = model_wrapper.predict(train_product_mols)
    check_results_against_targets(inf_results, train_reactant_smiles)

    # Run inference via the generic `get_results`.
    results_from_eval = get_results(model_wrapper, inputs=train_product_mols, num_results=10)
    check_results_against_targets(results_from_eval.results, train_reactant_smiles)



from dataclasses import dataclass, field
from typing import Any, Optional


@dataclass
class ChemEncoderConfig:
    fingerprint_dim: int = 16387  # Fingerprint dimension


@dataclass
class GNNEncoderConfig:
    # Hyperparams that control the output dimensions
    atom_out_channels: int = 128
    bond_out_channels: Optional[int] = None
    mol_out_channels: int = 512

    # Hyperparams for mapping categorical features to dense embeddings
    atom_categorical_features_channels: Optional[int] = None

    # Which GNN from `pytorch_geometric` to use, and with what hyperparameters
    gnn_class: str = "PNA"
    gnn_kwargs: dict[str, Any] = field(
        default_factory=lambda: dict(
            hidden_channels=64,
            num_layers=3,
            dropout=0.0,
            norm="GraphNorm",
            # Use Jumping Knowledge i.e. concatenate node-level outputs across all GNN layers:
            jk="cat",
            # Use all aggregators and scalers supported by the `PNAConv` layer:
            aggregators=["sum", "mean", "min", "max", "var", "std"],
            scalers=["identity", "amplification", "attenuation", "linear", "inverse_linear"],
            # Degree histogram used for normalization computed over products in USPTO-50K train set:
            deg=[4, 191634, 516869, 307522, 23296, 0, 15],
        )
    )

    # Aggregation hyperparameters
    aggregation_num_heads: int = 8
    aggregation_dropout: float = 0.5

    # Featurizer to map from raw inputs to graphs
    featurizer_class: str = "DGLLifeMoleculeFeaturizer"
    featurizer_kwargs: dict[str, Any] = field(default_factory=lambda: {})



"""Script for training a new model on a given dataset.

Currently supported model classes:
- TemplateClassificationChem: template classification on top of the `ChemEncoder`
- TemplateClassificationGNN: template classification on top of the `GNNEncoder`
- TemplateLocalization: template classification and localization using two `GNNEncoder`s
- SmilesTransformer: smiles based transformer models with better model flexibility and faster beam search
"""

import datetime
import math
import shutil
import sys
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Any, Optional, Union, cast

import pytorch_lightning as pl
import torch
from omegaconf import MISSING, OmegaConf
from syntheseus.reaction_prediction.utils.misc import set_random_seed

import retrochimera
from retrochimera.chem.rules import RuleBase
from retrochimera.data.processed_dataset import ProcessedDataModule
from retrochimera.encoders.configs import ChemEncoderConfig, GNNEncoderConfig
from retrochimera.models.lightning import AbstractModel
from retrochimera.models.smiles_transformer import SmilesTransformerModel
from retrochimera.models.template_classification import MCCModel
from retrochimera.models.template_localization import TemplateLocalizationModel
from retrochimera.utils.logging import get_logger
from retrochimera.utils.misc import convert_camel_to_snake, lookup_by_name
from retrochimera.utils.pytorch_lightning import ModelCheckpoint, OptLRMonitor
from retrochimera.utils.training import average_checkpoints

logger = get_logger(__name__)


HELP = """\
Train the model with a chosen preset and optional config overrides.

Required arguments:
  model_class         Specifies the model class to use. Must be one of 'TemplateClassificationChem',
                      'TemplateClassificationGNN', 'TemplateLocalization' or 'SmilesTransformer'.
  preset              Selects a configuration from 'pistachio' (default), 'uspto_full', 'uspto_50k'.
  processed_data_dir  Directory containing the processed data.
  checkpoint_dir      Directory where model checkpoints will be saved.

Optionally you can provide one or several paths to YAML config files to override the preset via the
'config' argument. Any config key can also be overridden directly from the CLI with key=value pairs.
"""


class ModelClass(Enum):
    TemplateClassificationChem = auto()
    TemplateClassificationGNN = auto()
    TemplateLocalization = auto()
    SmilesTransformer = auto()


@dataclass
class MLPConfig:
    """Stores hyperparameters configuring the MLP head."""

    n_layers: int = 0  # Number of extra hidden layers (not counting input and output layers)
    dropout: float = 0.2
    hidden_dim: int = 256


@dataclass
class TrainingConfig:
    """Stores hyperparameters controlling the training process."""

    batch_size: int = 128  # Batch size to use
    learning_rate: float = 1e-3  # Starting learning rate
    learning_rate_decay_step_size: int = 100  # Frequency (in epochs) for decaying the learning rate
    learning_rate_decay_rate: float = 0.1  # Coefficient for a single learning rate decay step
    optimizer_betas: tuple[float, float] = (0.9, 0.999)  # Momentum coefficients for Adam optimizer
    n_epochs: int = 100  # Maximum number of training epochs
    check_val_every_n_epoch: int = 1  # Frequency of running validation (in training epochs)
    num_checkpoints_for_averaging: int = 1  # How many best checkpoints to take an average of
    accelerator: str = "gpu"  # Whether to use a CPU, GPU or TPU
    n_devices: int = 1  # Number of devices
    gradient_clip_val: float = 50.0  # Maximum gradient norm to clip to.
    accumulate_grad_batches: int = 1  # Number of batches between gradient steps.


@dataclass
class TemplateClassificationChemConfig:
    """Configures the template classification model backed by the chem encoder."""

    encoder: ChemEncoderConfig = ChemEncoderConfig()
    mlp: MLPConfig = MLPConfig()
    label_smoothing: float = 0.0
    training: TrainingConfig = TrainingConfig()


@dataclass
class TemplateClassificationGNNConfig:
    """Configures the template classification model backed by the GNN encoder."""

    encoder: GNNEncoderConfig = GNNEncoderConfig()
    mlp: MLPConfig = MLPConfig(dropout=0.4, hidden_dim=128)
    label_smoothing: float = 0.0
    training: TrainingConfig = TrainingConfig(n_epochs=150)


@dataclass
class TemplateLocalizationConfig:
    """Configures the template localization model."""

    # Note: the encoders here do not have to be GNN-based, but they need to be able to provide
    # atom-level outputs, thus `ChemEncoder` would not suffice. The configuration below can be
    # generalized once we introduce other suitable encoders, e.g. based on graph transformers.

    input_encoder: GNNEncoderConfig = GNNEncoderConfig(
        aggregation_dropout=0.4
    )  # Encoder to use for input graphs
    rewrite_encoder: GNNEncoderConfig = GNNEncoderConfig(
        aggregation_dropout=0.4, featurizer_class="DGLLifeRewriteFeaturizer"
    )  # Encoder to use for rewrites
    classification_label_smoothing: float = 0.1
    localization_label_smoothing: float = 0.1
    classification_space_dim: Optional[int] = 256
    free_rewrite_embedding_dim: int = 0
    classification_loss_type: str = "softmax"  # can be either "softmax" or "sigmoid"
    classification_min_temperature: Optional[float] = None
    classification_max_temperature: Optional[float] = None
    negative_to_positive_targets_ratio: float = 0.0
    num_negative_rewrites_in_localization: int = 0
    rewrite_encoder_num_epochs: Optional[int] = None
    training: TrainingConfig = TrainingConfig(n_epochs=600, learning_rate_decay_step_size=550)


@dataclass
class SmilesTransformerConfig:
    """Configures the transformer model."""

    vocab_path: str = MISSING
    hidden_dim: int = 512
    n_layers: int = 6
    n_heads: int = 8
    feedforward_dim: int = 2048
    activation: str = "gelu"
    max_seq_len: int = 512
    dropout: float = 0.1
    label_smoothing: float = 0.0
    schedule: str = "cycle"
    warm_up_steps: int = 8000
    share_encoder_decoder_input_embedding: bool = True
    initialization: str = "xavier"
    add_qkvbias: bool = False
    layer_norm: str = "standard"
    num_kv: int = 0
    parallel_residual: bool = False
    training: TrainingConfig = TrainingConfig(n_epochs=200, batch_size=128)


@dataclass
class ModelTrainingConfig:
    """Base config for configuring a model to be trained."""

    model_class: ModelClass  # Which model type to train

    # Fields relevant to specific model types
    template_classification_chem_config: TemplateClassificationChemConfig = (
        TemplateClassificationChemConfig()
    )
    template_classification_gnn_config: TemplateClassificationGNNConfig = (
        TemplateClassificationGNNConfig()
    )
    template_localization_config: TemplateLocalizationConfig = TemplateLocalizationConfig()
    smiles_transformer_config: SmilesTransformerConfig = SmilesTransformerConfig()


@dataclass
class TrainConfig(ModelTrainingConfig):
    """Config for running training on a given dataset."""

    processed_data_dir: str = MISSING  # Directory for saving preprocessed data
    checkpoint_dir: str = MISSING  # Directory to store pytorch lightning model checkpoints
    log_dir: str = "."  # Directory to store TensorBoard logs
    seed: int = 0  # Seed to use for all sources of randomness (Python, numpy, torch)

    num_processes_training: int = 1  # Number of processes to use for training


@dataclass
class TrainingResult:
    model: AbstractModel
    test_results: list[dict[str, Any]]


def train(
    model: AbstractModel, data_path: Union[str, Path], config: TrainConfig, model_config: Any
) -> TrainingResult:
    data_loader_kwargs = {
        "batch_size": model_config.training.batch_size,
        "num_workers": config.num_processes_training,
        "collate_fn": model.collate,
        "persistent_workers": config.num_processes_training > 0,
        "pin_memory": True,
    }

    # Add a timeout if we are using data loading worker processes
    if data_loader_kwargs["num_workers"]:
        data_loader_kwargs["timeout"] = 600

    datamodule = ProcessedDataModule(h5_path=data_path, data_loader_kwargs=data_loader_kwargs)

    # Prepare a callback to save the checkpoint at the end of each epoch to recover from
    # preemptions; this is different from the callback that keeps track of the best checkpoint(s).
    last_checkpoint_callback = ModelCheckpoint(
        dirpath=config.checkpoint_dir,
        filename="{epoch}-last",
        every_n_epochs=1,
        save_on_train_epoch_end=True,
    )

    if isinstance(model, SmilesTransformerModel):
        batches_per_gpu = math.ceil(
            len(datamodule.train_dataloader()) / float(model_config.training.n_devices)
        )
        total_steps = (
            math.ceil(batches_per_gpu / model_config.training.accumulate_grad_batches)
            * model_config.training.n_epochs
        )
        model.total_steps = total_steps
        logger.info(f"Total training steps: {total_steps}")

        best_checkpoint_callback = ModelCheckpoint(
            dirpath=config.checkpoint_dir,
            filename="{epoch}-best",
            monitor="epoch",
            mode="max",
            save_top_k=model_config.training.num_checkpoints_for_averaging,
            every_n_epochs=1,
        )
        lr_monitor_callback = OptLRMonitor()

        callbacks = [best_checkpoint_callback, last_checkpoint_callback, lr_monitor_callback]
    else:
        best_checkpoint_callback = ModelCheckpoint(
            dirpath=config.checkpoint_dir,
            filename="{epoch}-best",
            monitor="val_mrr",
            save_top_k=model_config.training.num_checkpoints_for_averaging,
            mode="max",
        )

        callbacks = [best_checkpoint_callback, last_checkpoint_callback]

    checkpoint_dir = Path(config.checkpoint_dir)
    checkpoint_dir_contents = list(checkpoint_dir.iterdir())
    logger.info(f"Contents of the checkpoint directory: {checkpoint_dir_contents}")

    Path(config.log_dir).mkdir(parents=True, exist_ok=True)
    wandb_logger = pl.loggers.wandb.WandbLogger(save_dir=config.log_dir)

    if (
        isinstance(model, TemplateLocalizationModel)
        and config.template_localization_config.rewrite_encoder_num_epochs is not None
    ):
        # TODO(krmaziar): Handle this more directly.
        strategy_kwargs = {"find_unused_parameters": True}
    else:
        strategy_kwargs = {}

    trainer = pl.Trainer(
        strategy=pl.strategies.DDPStrategy(
            timeout=datetime.timedelta(seconds=36000), **strategy_kwargs
        ),  # Increase timeout to give nodes enough time to load the rulebase
        enable_checkpointing=True,
        max_epochs=model_config.training.n_epochs,
        check_val_every_n_epoch=model_config.training.check_val_every_n_epoch,
        logger=wandb_logger,
        callbacks=callbacks,
        accelerator=model_config.training.accelerator,
        devices=model_config.training.n_devices,
        gradient_clip_val=model_config.training.gradient_clip_val,
        accumulate_grad_batches=model_config.training.accumulate_grad_batches,
        num_sanity_val_steps=0,
    )

    best_checkpoint_path = checkpoint_dir / "combined.ckpt"

    if best_checkpoint_path.exists():
        logger.info("Found a combined checkpoint which means training has already completed")
    else:
        last_checkpoint_paths = [
            path for path in checkpoint_dir_contents if str(path).endswith("-last.ckpt")
        ]
        if len(last_checkpoint_paths) > 1:
            logger.warning("Ambiguous state of the checkpoint directory")
            last_checkpoint_path = max(
                last_checkpoint_paths,
                key=lambda path: int(str(path).removesuffix("-last.ckpt").split("=")[-1]),
            )
        elif len(last_checkpoint_paths) == 1:
            [last_checkpoint_path] = last_checkpoint_paths
        else:
            last_checkpoint_path = None

        if last_checkpoint_path is not None:
            logger.info(f"Resuming from {last_checkpoint_path}")
            fit_kwargs = {"ckpt_path": last_checkpoint_path}
        else:
            logger.info("No checkpoint found, starting training from scratch")
            fit_kwargs = {}

        trainer.fit(model, datamodule=datamodule, **fit_kwargs)

        if trainer.global_rank == 0:
            checkpoint_paths_for_averaging = list(best_checkpoint_callback.best_k_models.keys())
            logger.info(
                f"Averaging out the following checkpoints: {checkpoint_paths_for_averaging}"
            )

            average_checkpoints(
                input_paths=checkpoint_paths_for_averaging, output_path=best_checkpoint_path
            )

    logger.info(f"Best checkpoint: {best_checkpoint_path}")

    last_checkpoints_dir = checkpoint_dir / "last"
    best_checkpoints_dir = checkpoint_dir / "best"

    model.eval()

    if last_checkpoints_dir.exists() and best_checkpoints_dir.exists():
        logger.info("Judging from checkpoint directory organization testing was already done")
        test_results = []
    else:
        test_results = trainer.test(ckpt_path=best_checkpoint_path, datamodule=datamodule)

    if trainer.global_rank == 0:
        last_checkpoints_dir.mkdir(exist_ok=True)
        best_checkpoints_dir.mkdir(exist_ok=True)

        for path in checkpoint_dir.iterdir():
            if str(path).endswith(".ckpt"):
                if str(path).endswith("-last.ckpt"):
                    logger.info(f"Moving {path} under {last_checkpoints_dir}")
                    path.rename(last_checkpoints_dir / path.name)
                elif path.name != best_checkpoint_path.name:
                    logger.info(f"Moving {path} under {best_checkpoints_dir}")
                    path.rename(best_checkpoints_dir / path.name)
        logger.info(
            f"Contents of the checkpoint directory after training: {list(checkpoint_dir.iterdir())}"
        )

    return TrainingResult(model=model, test_results=test_results)


def build_model_from_config(
    config: ModelTrainingConfig, rulebase: RuleBase, rulebase_dir: Union[str, Path]
) -> tuple[AbstractModel, Any]:
    model_class: Any = None
    model_config: Any = None
    model_kwargs: dict[str, Any] = {}

    def add_encoder_kwargs(
        prefix: str,
        encoder_class: str,
        encoder_config: Union[ChemEncoderConfig, GNNEncoderConfig],
    ) -> None:
        model_kwargs[f"{prefix}encoder_class"] = encoder_class
        model_kwargs[f"{prefix}encoder_kwargs"] = OmegaConf.to_container(encoder_config)

    if config.model_class == ModelClass.TemplateClassificationChem:
        model_class = MCCModel
        model_config = config.template_classification_chem_config

        add_encoder_kwargs(
            prefix="", encoder_class="ChemEncoder", encoder_config=model_config.encoder
        )
    elif config.model_class == ModelClass.TemplateClassificationGNN:
        model_class = MCCModel
        model_config = config.template_classification_gnn_config

        add_encoder_kwargs(
            prefix="", encoder_class="GNNEncoder", encoder_config=model_config.encoder
        )
    elif config.model_class == ModelClass.TemplateLocalization:
        model_class = TemplateLocalizationModel
        model_config = config.template_localization_config

        add_encoder_kwargs(
            prefix="input_", encoder_class="GNNEncoder", encoder_config=model_config.input_encoder
        )
        add_encoder_kwargs(
            prefix="rewrite_",
            encoder_class="GNNEncoder",
            encoder_config=model_config.rewrite_encoder,
        )

        rewrite_featurizer_class = lookup_by_name(
            retrochimera.encoders.featurizers,
            model_kwargs["rewrite_encoder_kwargs"]["featurizer_class"],
        )

        # Look through the rulebase to build the featurizer kwargs (e.g. atom SMARTS vocabulary).
        rewrite_featurizer_kwargs = rewrite_featurizer_class.prepare_kwargs(
            rewrites=(rule.rxn for rule in rulebase.rules.values())
        )
        model_kwargs["rewrite_encoder_kwargs"]["featurizer_kwargs"] = rewrite_featurizer_kwargs
    elif config.model_class == ModelClass.SmilesTransformer:
        model_class = SmilesTransformerModel
        model_config = config.smiles_transformer_config
    else:
        raise ValueError(f"Model class {config.model_class} not recognized")

    # Add model kwargs that are shared across all models classes.
    model_kwargs.update(
        {
            "n_classes": len(rulebase),
            "learning_rate": model_config.training.learning_rate,
            "learning_rate_decay_step_size": model_config.training.learning_rate_decay_step_size,
            "learning_rate_decay_rate": model_config.training.learning_rate_decay_rate,
            "optimizer_betas": list(model_config.training.optimizer_betas),
        }
    )

    if config.model_class == ModelClass.TemplateLocalization:
        keys = [
            "classification_label_smoothing",
            "localization_label_smoothing",
            "classification_space_dim",
            "free_rewrite_embedding_dim",
            "classification_loss_type",
            "classification_min_temperature",
            "classification_max_temperature",
            "negative_to_positive_targets_ratio",
            "num_negative_rewrites_in_localization",
            "rewrite_encoder_num_epochs",
        ]
        model_kwargs.update({key: getattr(model_config, key) for key in keys})
        model_kwargs["num_total_rewrite_lhs_atoms"] = sum(
            rule.rxn.rdkit_lhs_mol.GetNumAtoms() for rule in rulebase.rules.values()
        )
    elif config.model_class == ModelClass.SmilesTransformer:
        keys = [
            "vocab_path",
            "hidden_dim",
            "n_layers",
            "n_heads",
            "feedforward_dim",
            "activation",
            "max_seq_len",
            "dropout",
            "label_smoothing",
            "schedule",
            "warm_up_steps",
            "share_encoder_decoder_input_embedding",
            "initialization",
            "add_qkvbias",
            "layer_norm",
            "num_kv",
            "parallel_residual",
        ]
        model_kwargs.update({key: getattr(model_config, key) for key in keys})
    else:
        model_kwargs.update(
            {
                "hidden_dim": model_config.mlp.hidden_dim,
                "n_hidden_layers": model_config.mlp.n_layers,
                "dropout": model_config.mlp.dropout,
            }
        )

    if config.model_class in [
        ModelClass.TemplateClassificationChem,
        ModelClass.TemplateClassificationGNN,
    ]:
        model_kwargs["label_smoothing"] = model_config.label_smoothing

    logger.info(f"Building the model {config.model_class} with kwargs {model_kwargs}")

    model = model_class(**model_kwargs)
    model.set_rulebase(rulebase=rulebase, rulebase_dir=rulebase_dir)

    return model, model_config


def parse_training_config(argv: list[str]) -> TrainConfig:
    config = OmegaConf.from_dotlist(argv)
    preset = config.pop("preset", "pistachio")

    # Use preset and model class to determine the config file to load.
    config_paths = [
        Path(__file__).parent
        / "config"
        / preset
        / (convert_camel_to_snake(config.model_class) + ".yaml")
    ]

    if "config" in config:
        config_paths.append(config.pop("config"))

    combined_config = OmegaConf.merge(
        OmegaConf.structured(TrainConfig), *[OmegaConf.load(c) for c in config_paths], config
    )
    return cast(TrainConfig, combined_config)


def main() -> None:
    torch.multiprocessing.set_sharing_strategy("file_system")

    argv = sys.argv[1:]

    # If no arguments were provided or `--help`/`-h` is present, print the help message and exit.
    if not argv or "--help" in argv or "-h" in argv:
        print(HELP)
        sys.exit(0)

    config = parse_training_config(argv)
    logger.info(f"Running training with the following config: {config}")

    set_random_seed(config.seed)

    processed_data_dir = Path(config.processed_data_dir)
    processed_rulebase_path = processed_data_dir / RuleBase.DEFAULT_FILE_NAME

    checkpoint_dir = Path(config.checkpoint_dir)
    checkpoint_dir.mkdir(parents=True, exist_ok=True)
    checkpoint_rulebase_path = checkpoint_dir / RuleBase.DEFAULT_FILE_NAME

    # Copy the rulebase into the checkpoint directory, which is only done by the first process
    # (other training processes will be spawned in `train`).
    if not checkpoint_rulebase_path.exists():
        logger.info(f"Copying the rulebase into {config.checkpoint_dir}")
        shutil.copy(processed_rulebase_path, checkpoint_rulebase_path)

    if config.model_class is ModelClass.SmilesTransformer:
        checkpoint_vocab_path = checkpoint_dir / SmilesTransformerModel.DEFAULT_VOCAB_FILE_NAME

        if not checkpoint_vocab_path.exists():
            # For the SMILES Transformer we also copy the vocab file.
            logger.info(f"Copying the vocab file into {config.checkpoint_dir}")
            shutil.copy(config.smiles_transformer_config.vocab_path, checkpoint_vocab_path)

    model, model_config = build_model_from_config(
        config=config,
        rulebase=RuleBase.load_from_file(dir=checkpoint_dir),
        rulebase_dir=checkpoint_dir,
    )

    logger.info("Starting training")
    train(
        model=model,
        data_path=str(processed_data_dir / "data.h5"),
        config=config,
        model_config=model_config,
    )


if __name__ == "__main__":
    main()



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/modules/average_attn.py
"""
from typing import Optional

import torch
import torch.nn as nn
from torch import Tensor

from retrochimera.opennmt.modules.position_ffn import ActivationFunction, PositionwiseFeedForward


def cumulative_average_mask(
    batch_size: int, t_len: int, device: Optional[torch.device] = None
) -> Tensor:
    """
    Builds the mask to compute the cumulative average as described in
    :cite:`DBLP:journals/corr/abs-1805-00631` -- Figure 3
    Args:
        batch_size (int): batch size
        t_len (int): length of the layer_in

    Returns:
        (Tensor):
        * A Tensor of shape ``(batch_size, t_len, t_len)``
    """

    triangle = torch.tril(torch.ones(t_len, t_len, dtype=torch.float, device=device))
    weights = torch.ones(1, t_len, dtype=torch.float, device=device) / torch.arange(
        1, t_len + 1, dtype=torch.float, device=device
    )
    mask = triangle * weights.transpose(0, 1)

    return mask.unsqueeze(0).expand(batch_size, t_len, t_len)


def cumulative_average(layer_in: Tensor, layer_cache: tuple, mask=None, step=None) -> Tensor:
    """
    Computes the cumulative average as described in
    :cite:`DBLP:journals/corr/abs-1805-00631` -- Equations (1) (5) (6)

    Args:
        layer_in (FloatTensor): sequence to average
            ``(batch_size, input_len, dimension)``
        layer_cache: tuple(bool, dict)
        if layer_cahe[0] is True use step otherwise mask
        mask: mask matrix used to compute the cumulative average
        step: current step of the dynamic decoding

    Returns:
        a tensor of the same shape and type as ``layer_in``.
    """

    if layer_cache[0]:
        average_attention = (layer_in + step * layer_cache[1]["prev_g"]) / (step + 1)
        layer_cache[1]["prev_g"] = average_attention
        return average_attention
    else:
        return torch.matmul(mask.to(layer_in.dtype), layer_in)


class AverageAttention(nn.Module):
    # class AverageAttention(torch.jit.ScriptModule):
    """
    Average Attention module from
    "Accelerating Neural Transformer via an Average Attention Network"
    :cite:`DBLP:journals/corr/abs-1805-00631`.

    Args:
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
       pos_ffn_activation_fn (ActivationFunction):
           activation function choice for PositionwiseFeedForward layer
    """

    def __init__(
        self,
        model_dim,
        dropout=0.1,
        aan_useffn=False,
        pos_ffn_activation_fn=ActivationFunction.relu,
    ):
        self.model_dim = model_dim
        self.aan_useffn = aan_useffn
        super(AverageAttention, self).__init__()
        if aan_useffn:
            self.average_layer = PositionwiseFeedForward(
                model_dim, model_dim, dropout, pos_ffn_activation_fn
            )
        self.gating_layer = nn.Linear(model_dim * 2, model_dim * 2)
        self.layer_cache = False, {"prev_g": torch.tensor([])}

    # @torch.jit.script
    def forward(self, layer_in, mask=None, step=None):
        """
        Args:
            layer_in (FloatTensor): ``(batch, t_len, dim)``

        Returns:
            (FloatTensor, FloatTensor):

            * gating_out ``(batch, tlen, dim)``
            * average_out average attention
                ``(batch, input_len, dim)``
        """

        batch_size = layer_in.size(0)
        t_len = layer_in.size(1)
        mask = (
            cumulative_average_mask(batch_size, t_len, layer_in.device)
            if not self.layer_cache[0]
            else None
        )
        average_out = cumulative_average(layer_in, self.layer_cache, mask, step)
        if self.aan_useffn:
            average_out = self.average_layer(average_out)
        gating_out = self.gating_layer(torch.cat((layer_in, average_out), -1))
        input_gate, forget_gate = torch.chunk(gating_out, 2, dim=2)
        gating_out = torch.sigmoid(input_gate) * layer_in + torch.sigmoid(forget_gate) * average_out

        return gating_out, average_out



import itertools
import json
from collections import defaultdict
from pathlib import Path
from typing import Any, Optional, Sequence, Union

import omegaconf
import torch
from syntheseus import BackwardReactionModel, Bag, Molecule, SingleProductReaction
from syntheseus.interface.reaction import ReactionMetaData
from syntheseus.reaction_prediction.inference_base import ExternalBackwardReactionModel

from retrochimera import inference
from retrochimera.utils.misc import lookup_by_name


class RetroChimeraModel(ExternalBackwardReactionModel):
    """Wrapper for the RetroChimera model."""

    def __init__(
        self,
        *args,
        model_dir: Optional[Union[str, Path]] = None,
        probability_from_score_temperature: float = 8.0,
        **kwargs,
    ) -> None:
        """Initializes the RetroChimera model wrapper.

        Assumed format of the model directory:
        - `model_dir` contains the ensembling config in the `models.json` file
        - `model_dir/[SUBMODEL]` contains the checkpoint files for model `SUBMODEL`
        """

        model_dir = Path(model_dir or self.get_default_model_dir())
        model_data = self._load_model_data_from_dir(Path(model_dir))

        model_kwargs = {}
        for key in model_data:
            if key in kwargs:
                # If the RetroChimera constructor receives an argument whose name matches one of the
                # constituent models, pass through the value as model kwargs downstream.
                value = kwargs[key]
                if not isinstance(value, (dict, omegaconf.DictConfig)):
                    raise RuntimeError(
                        f"Value for {key} should be the model kwargs for {model_data[key][0]}; "
                        f"found value of type {type(value)} instead of dict"
                    )

                model_kwargs[key] = value
                del kwargs[key]

        # Having removed kwargs meant for constituent models we can now call base class `__init__`.
        super().__init__(*args, model_dir=model_dir, **kwargs)

        self._init_from_dir(model_dir=model_dir, model_data=model_data, model_kwargs=model_kwargs)
        self.probability_from_score_temperature = probability_from_score_temperature

    def _load_model_data_from_dir(self, model_dir: Path) -> dict[str, tuple[str, list[float]]]:
        with open(model_dir / "models.json") as f:
            return json.load(f)

    def _init_from_dir(
        self,
        model_dir: Path,
        model_data: dict[str, tuple[str, list[float]]],
        model_kwargs: dict[str, Any],
    ) -> None:
        self._models: list[BackwardReactionModel] = []
        self._model_names: list[str] = []
        self._model_weights: list[list[float]] = []

        for model_subdir, (model_class_name, model_weights) in model_data.items():
            model_class = lookup_by_name(inference, model_class_name)

            self._models.append(
                model_class(
                    model_dir=model_dir / model_subdir,
                    device=self.device,
                    **model_kwargs.get(model_subdir, {}),
                )
            )
            self._model_names.append(model_subdir)
            self._model_weights.append(model_weights)

    def get_parameters(self):
        return itertools.chain(*(model.get_parameters() for model in self._models))

    def _get_reactions(
        self, inputs: list[Molecule], num_results: int
    ) -> list[Sequence[SingleProductReaction]]:
        model_batch_results: list[list[Sequence[SingleProductReaction]]] = [
            model(inputs=inputs, num_results=num_results) for model in self._models
        ]

        return [
            combine_results(
                input=input,
                model_results=list(model_results),
                model_weights=self._model_weights,
                num_results=num_results,
                model_names=self._model_names,
                probability_from_score_temperature=self.probability_from_score_temperature,
            )
            for input, model_results in zip(inputs, zip(*model_batch_results))
        ]


def combine_results(
    input: Molecule,
    model_results: list[Sequence[SingleProductReaction]],
    model_weights: list[list[float]],
    num_results: int,
    model_names: Optional[list[str]] = None,
    probability_from_score_temperature: Optional[float] = None,
) -> list[SingleProductReaction]:
    # Generate default model names if not provided.
    model_names = model_names or [f"model_{i}" for i in range(len(model_results))]

    result_to_ranks: dict[Bag[Molecule], list[Optional[int]]] = defaultdict(
        lambda: [None] * len(model_results)
    )
    result_to_metadata: dict[Bag[Molecule], list[Optional[ReactionMetaData]]] = defaultdict(
        lambda: [None] * len(model_results)
    )

    for model_idx, results in enumerate(model_results):
        for rank, r in enumerate(results):
            result_to_ranks[r.reactants][model_idx] = rank
            result_to_metadata[r.reactants][model_idx] = r.metadata

    result_to_score: dict[Bag[Molecule], float] = {}
    for result, ranks in result_to_ranks.items():
        result_to_score[result] = 0.0
        for rk, weights in zip(ranks, model_weights):
            if rk is not None:
                if rk < len(weights):
                    result_to_score[result] += weights[rk]
                else:
                    # Extend the weights sequence at the end for tie-breaking.
                    result_to_score[result] += weights[-1] / (rk - len(weights) + 2)

    combined_results = sorted(
        result_to_score.keys(), key=lambda r: result_to_score[r], reverse=True
    )[:num_results]

    metadata_list = []
    for rct in combined_results:
        metadata_list.append(
            ReactionMetaData(  # type: ignore[typeddict-unknown-key]
                score=result_to_score[rct],
                individual_ranks={
                    model_name: result_to_ranks[rct][idx]
                    for idx, model_name in enumerate(model_names)
                },
                individual_metadata={
                    model_name: result_to_metadata[rct][idx]
                    for idx, model_name in enumerate(model_names)
                },
            )
        )

    if probability_from_score_temperature is not None:
        # Divide the scores by the sum of rank-1 scores across all models (so that the best possible
        # score is normalized to 1), then multiply by the given temperature before applying softmax.
        factor = probability_from_score_temperature / sum(weights[0] for weights in model_weights)
        scores = [result_to_score[r] * factor for r in combined_results]
        probabilities = torch.nn.functional.softmax(torch.as_tensor(scores), dim=-1)

        for metadata, probability in zip(metadata_list, probabilities.numpy().tolist()):
            metadata["probability"] = probability

    return [
        SingleProductReaction(product=input, reactants=reactants, metadata=metadata)
        for reactants, metadata in zip(combined_results, metadata_list)
    ]



from rdkit import Chem
from rdkit.Chem import AllChem
from syntheseus.reaction_prediction.chem.utils import ATOM_MAPPING_PROP_NAME


def get_mapped_atoms(mol: Chem.Mol) -> list[Chem.Atom]:
    """Extract all atoms with atom mapping.

    Args:
        mol: Molecule to extract the atoms from.

    Returns:
        List of atoms that have atom mapping.
    """
    return [atom for atom in mol.GetAtoms() if atom.HasProp(ATOM_MAPPING_PROP_NAME)]


def get_atom_mapping(mol: Chem.Mol) -> dict[int, int]:
    """Extract a mapping from atom mapping ID to atom ID.

    Args:
        mol: Molecule to extract the atom mapping from.

    Returns:
        Dictionary mapping IDs assigned under the atom mapping to RDKit atom IDs.
    """
    atom_mapping_num_to_idx = {}

    for atom in get_mapped_atoms(mol):
        atom_mapping_num_to_idx[atom.GetAtomMapNum()] = atom.GetIdx()

    return atom_mapping_num_to_idx


def get_all_atom_mapping_nums(mol: Chem.Mol) -> list[int]:
    """Extract all atom mapping numbers.

    Args:
        mol: Molecule to extract the numbers from.

    Returns:
        List of atom mapping IDs.
    """
    return [atom.GetAtomMapNum() for atom in get_mapped_atoms(mol)]


def normalize_atom_mapping(reaction: AllChem.ChemicalReaction) -> None:
    """Normalize atom mapping in a reaction in-place to use `[1, 2, ...]`."""

    def get_mapped_atoms(mol: Chem.Mol) -> list[Chem.Atom]:
        return [
            (atom, atom.GetAtomMapNum())
            for atom in mol.GetAtoms()
            if atom.HasProp(ATOM_MAPPING_PROP_NAME)
        ]

    reactant_mapped_atoms: list[Chem.Atom] = sum(
        [get_mapped_atoms(mol) for mol in reaction.GetReactants()], []
    )
    product_mapped_atoms: list[Chem.Atom] = sum(
        [get_mapped_atoms(mol) for mol in reaction.GetProducts()], []
    )

    reactant_atom_ids = set(id for _, id in reactant_mapped_atoms)
    product_atom_ids = set(id for _, id in product_mapped_atoms)

    if len(reactant_atom_ids) < len(reactant_mapped_atoms):
        raise ValueError("Reactant side of the reaction contains duplicate mapping IDs")

    if len(product_atom_ids) < len(product_mapped_atoms):
        raise ValueError("Product side of the reaction contains duplicate mapping IDs")

    id_mapping = {
        id_old: id_new + 1
        for id_new, id_old in enumerate(sorted(reactant_atom_ids & product_atom_ids))
    }

    for atom, id in reactant_mapped_atoms + product_mapped_atoms:
        if id in id_mapping:
            atom.SetAtomMapNum(id_mapping[id])
        else:
            atom.ClearProp(ATOM_MAPPING_PROP_NAME)


def map_removed_lhs_atoms_to_dummies(
    reaction: AllChem.ChemicalReaction,
) -> tuple[AllChem.ChemicalReaction, int]:
    """Add a dummy atom to the right-hand side for each unmapped atom on the left-hand side.

    Returns:
        A modified reaction and the largest atom mapping ID that is *not* mapped to a dummy atom.
    """

    reactant_unmapped_atoms: list[Chem.Atom] = []
    max_atom_mapping_num = -1

    for mol in reaction.GetReactants():
        for atom in mol.GetAtoms():
            if atom.HasProp(ATOM_MAPPING_PROP_NAME):
                max_atom_mapping_num = max(max_atom_mapping_num, atom.GetAtomMapNum())
            else:
                reactant_unmapped_atoms.append(atom)

    reactants = list(reaction.GetReactants())
    products = list(reaction.GetProducts())

    product_new = Chem.RWMol(products[0])
    curr_atom_mapping_num = max_atom_mapping_num + 1

    for atom in reactant_unmapped_atoms:
        atom.SetAtomMapNum(curr_atom_mapping_num)

        new_atom = Chem.Atom(atom.GetSymbol())
        new_atom.SetAtomMapNum(curr_atom_mapping_num)
        product_new.AddAtom(new_atom)

        curr_atom_mapping_num += 1

    products[0] = Chem.Mol(product_new)

    # Create an empty reaction and add back all the reactant/product templates.
    reaction_new = AllChem.ReactionFromSmarts(">>")

    for reactant in reactants:
        reaction_new.AddReactantTemplate(reactant)

    for product in products:
        reaction_new.AddProductTemplate(product)

    return reaction_new, max_atom_mapping_num



from rdkit import Chem
from syntheseus.interface.molecule import Molecule

from retrochimera.chem.rewrite import Rewrite
from retrochimera.encoders.featurizers import (
    DGLLifeMoleculeFeaturizer,
    DGLLifeRewriteFeaturizer,
    RawGraph,
    RawRewriteGraph,
)


def test_molecule_featurizer() -> None:
    featurizer = DGLLifeMoleculeFeaturizer.get_default()
    molecule = Molecule("COCCOc1ccc(C(C)=O)c(O)c1")

    raw_graph = featurizer(molecule)
    assert isinstance(raw_graph, RawGraph)
    assert not isinstance(raw_graph, RawRewriteGraph)

    rdkit_mol = Chem.MolFromSmiles(molecule.smiles)

    assert raw_graph.atom_features.shape == (rdkit_mol.GetNumAtoms(), featurizer.num_atom_features)
    assert raw_graph.bond_features.shape == (len(raw_graph.bonds), featurizer.num_bond_features)
    assert raw_graph.bonds.shape == (2 * rdkit_mol.GetNumBonds(), 2)


def test_rewrite_featurizer() -> None:
    featurizer = DGLLifeRewriteFeaturizer.get_default(
        atom_smarts_vocab=["C", "O", "[C&H1&+0&D2]", "[O&H0&+0&D1]", "[C&H1&+0&D3]", "[O&H0&+0&D2]"]
    )
    rewrite = Rewrite.from_rdkit(
        "[C:1]-[C&H1&+0&D2:3]=[O&H0&+0&D1:2]>>[C:1]-[C&H1&+0&D3:3]1-[O&H0&+0&D2:2]-C-C-O-1"
    )

    raw_graph = featurizer(rewrite)
    assert isinstance(raw_graph, RawRewriteGraph)

    rdkit_lhs = Chem.MolFromSmarts(rewrite.lhs)
    rdkit_rhs = Chem.MolFromSmarts(rewrite.rhs)
    rdkit_total_num_atoms = rdkit_lhs.GetNumAtoms() + rdkit_rhs.GetNumAtoms()
    rdkit_total_num_bonds = rdkit_lhs.GetNumBonds() + rdkit_rhs.GetNumBonds()

    assert raw_graph.atom_features.shape == (rdkit_total_num_atoms, featurizer.num_atom_features)
    assert raw_graph.bond_features.shape == (len(raw_graph.bonds), featurizer.num_bond_features)
    assert raw_graph.bonds.shape == (2 * (rdkit_total_num_bonds + len(rewrite.mapping)), 2)
    assert raw_graph.node_in_lhs.shape == (rdkit_total_num_atoms,)



from rdkit import Chem
from rdkit.Chem import AllChem

from retrochimera.chem.descriptors import (
    calc_ap,
    calc_morgan,
    calc_rdkfp,
    calc_rxn_fp,
    calc_rxn_morgan,
    fp_to_dense_np,
    fp_to_sparse_array,
    tanimoto,
    tanimoto_sparse,
)


def test_fp_to_dense() -> None:
    mol = Chem.MolFromSmiles("CCCCOCCCF")

    fpr = calc_morgan(mol)
    a = fp_to_dense_np(fp=fpr, size=128)
    assert a.sum() >= 0


def test_fp_to_sparse() -> None:
    mol = Chem.MolFromSmiles("CCCCOCCCF")

    fpr = calc_morgan(mol)
    a = fp_to_sparse_array(fp=fpr, size=128)
    assert a.sum() >= 0


def test_rdkit_rxnfp() -> None:
    r1 = AllChem.ReactionFromSmarts("OCCCO>>FCCCF")
    r2 = AllChem.ReactionFromSmarts("FOOF.CCCCl>>CCCO")
    r3 = AllChem.ReactionFromSmarts("OCCCCO>>FCCCCF")

    fpr1 = calc_rxn_morgan(r1)
    fpr2 = calc_rxn_morgan(r2)
    fpr3 = calc_rxn_morgan(r3)

    a = fp_to_dense_np(fp=fpr1, size=128)
    b = fp_to_dense_np(fp=fpr2, size=128)
    c = fp_to_dense_np(fp=fpr3, size=128)

    assert a.sum() == 0
    assert b.sum() != 0

    res = tanimoto(a.reshape(1, -1), c.reshape(1, -1))
    assert res.shape == (1,)
    assert 0.0 < res.item() < 1.0


def test_custom_rxnfp() -> None:
    r1 = AllChem.ReactionFromSmarts(
        "OC(=O)C1=CC=CC=C1.C1CCNC1>>O=C(N1CCCC1)C1=CC=CC=C1", useSmiles=1
    )
    r2 = AllChem.ReactionFromSmarts("FOOF.CCCCl>>CCCO", useSmiles=1)
    r3 = AllChem.ReactionFromSmarts(
        "CC1=CC=CC=C1C(O)=O.C1CCNC1>>CC1=C(C=CC=C1)C(=O)N1CCCC1", useSmiles=1
    )

    a = calc_rxn_fp(r1, fp_fn=calc_morgan)
    b = calc_rxn_fp(r2, fp_fn=calc_morgan)
    c = calc_rxn_fp(r3, fp_fn=calc_morgan)

    assert b.sum() != 0
    res = tanimoto_sparse(a.reshape(1, -1), c.reshape(1, -1))
    assert res.shape == (1,)
    assert 0.0 < res.item() < 1.0


def test_custom_ap_rxnfp() -> None:
    r1 = AllChem.ReactionFromSmarts(
        "OC(=O)C1=CC=CC=C1.C1CCNC1>>O=C(N1CCCC1)C1=CC=CC=C1", useSmiles=1
    )
    r2 = AllChem.ReactionFromSmarts(
        "CC1=CC=CC=C1C(O)=O.C1CCNC1>>CC1=C(C=CC=C1)C(=O)N1CCCC1", useSmiles=1
    )

    a = calc_rxn_fp(r1, fp_fn=calc_ap, size=1000)
    b = calc_rxn_fp(r2, fp_fn=calc_ap, size=1000)

    res = tanimoto_sparse(a, b)
    assert res.shape == (1,)
    assert 0.0 < res.item() < 1.0


def test_custom_rdk_rxnfp() -> None:
    r1 = AllChem.ReactionFromSmarts(
        "OC(=O)C1=CC=CC=C1.C1CCNC1>>O=C(N1CCCC1)C1=CC=CC=C1", useSmiles=1
    )
    r2 = AllChem.ReactionFromSmarts(
        "CC1=CC=CC=C1C(O)=O.C1CCNC1>>CC1=C(C=CC=C1)C(=O)N1CCCC1", useSmiles=1
    )

    a = calc_rxn_fp(r1, fp_fn=calc_rdkfp)
    b = calc_rxn_fp(r2, fp_fn=calc_rdkfp)

    res = tanimoto_sparse(a, b)
    assert res.shape == (1,)
    assert 0.0 < res.item() < 1.0



from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Callable, Generic, Iterable, Union

from syntheseus.interface.molecule import Molecule
from syntheseus.reaction_prediction.utils.misc import parallelize

from retrochimera.chem.rules import RuleBase
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.encoders.base import DataType, Encoder


def get_unique_template_id(sample: TemplateReactionSample) -> int:
    ground_truth_results = [
        result for result in sample.template_application_results if result.ground_truth
    ]

    if len(ground_truth_results) != 1:
        raise ValueError(f"Warning: incorrect number of ground-truth results in {sample}")

    return ground_truth_results[0].template_id


@dataclass
class ProcessedSample(Generic[DataType]):
    input: DataType  # Featurized input molecule
    target: int  # ID of the ground-truth rewrite


def _process_sample(
    sample: tuple[Molecule, int], input_preprocess_fn: Callable[[Molecule], DataType]
) -> ProcessedSample:
    input, target = sample
    return ProcessedSample(input=input_preprocess_fn(input), target=target)


def preprocess_samples(
    samples: Iterable[TemplateReactionSample],
    rulebase_dir: Union[str, Path],
    encoder: Encoder,
    num_processes: int,
) -> Iterable[ProcessedSample]:
    # Extract the relevant info from each sample.
    samples_processed = (
        (Molecule(smiles=sample.products_str), get_unique_template_id(sample)) for sample in samples
    )

    rule_ids = RuleBase.load_rule_ids_from_file(dir=rulebase_dir)

    # Get rid of labels which are not in the `rulebase` (e.g. they were too rare).
    samples_processed = (sample for sample in samples_processed if sample[-1] in rule_ids)

    yield from parallelize(
        partial(_process_sample, input_preprocess_fn=encoder.preprocess),
        samples_processed,
        num_processes=num_processes,
    )



import dataclasses
from collections import Counter
from dataclasses import dataclass
from hashlib import md5
from typing import Any, Iterable, Optional, cast

import joblib
from rdchiral import template_extractor
from syntheseus.reaction_prediction.chem.utils import (
    molecule_bag_from_smiles_strict,
    remove_atom_mapping,
)
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from syntheseus.reaction_prediction.utils.misc import cpu_count, set_random_seed

from retrochimera.chem.rules import RuleBase
from retrochimera.data.dataset import DataFold, DiskReactionDataset
from retrochimera.data.template_reaction_sample import (
    TemplateApplicationResult,
    TemplateReactionSample,
)
from retrochimera.data.utils import load_raw_reactions_files
from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)


@dataclass
class ExtractTemplatesConfig:
    """Config for extracting templates from a dataset."""

    data_dir: str  # Directory containing raw data.
    output_dir: str  # Directory to write output to.
    min_freq: int = 0  # Minimum template frequency in training data for inclusion into the library.
    num_processes: int = cpu_count()  # Number of parallel processes to use.


def _convert(rxn_smiles: str, rxn_id: int) -> dict[str, Any]:
    """Convert into rdchiral format.

    Args:
        rxn_smiles: Mapped reaction SMILES.
        rxn_id: ID of the reaction.

    Returns:
        Dict of the form {"reactants": reactants, "products": prod, "_id": rxn_id}.
    """

    reactants, _, products = rxn_smiles.split(">")
    return {"reactants": reactants, "products": products, "_id": rxn_id}


def extract(rxn_smiles: str, rxn_id: int, wrap_smarts: bool) -> Optional[TemplateReactionSample]:
    # For Pistachio we may have extra info after the reaction SMILES, so get rid of it first.
    rxn_smiles = rxn_smiles.split("\t")[0].split(" ")[0]
    rdch = _convert(rxn_smiles, rxn_id)

    try:
        template = template_extractor.extract_from_reaction(rdch)["reaction_smarts"]

        if wrap_smarts:
            reactants, conditions, products = template.split(">")
            template = f"({reactants})>{conditions}>({products})"

        return TemplateReactionSample(
            reactants=molecule_bag_from_smiles_strict(remove_atom_mapping(rdch["reactants"])),
            products=molecule_bag_from_smiles_strict(remove_atom_mapping(rdch["products"])),
            mapped_reaction_smiles=rxn_smiles,
            template=template,
        )
    except Exception:
        return None


def extract_templates(
    rxn_smiles_list: list[str], wrap_smarts: bool, num_processes: int = cpu_count()
) -> list[TemplateReactionSample]:
    pool = joblib.Parallel(n_jobs=num_processes)
    jobs = (joblib.delayed(extract)(s, i, wrap_smarts) for i, s in enumerate(rxn_smiles_list))

    return [result for result in pool(jobs) if result is not None]


@dataclass
class LabeledMolsAndTemplates:
    samples: dict[DataFold, Iterable[TemplateReactionSample]]
    rulebase: RuleBase


def rulebase_from_templates_list(
    templates: list[str], min_template_occurrence: int = 0
) -> tuple[RuleBase, dict[str, int]]:
    # Count how often templates appear
    counter = Counter(templates)

    rulebase = RuleBase()
    for template, count in counter.most_common(len(counter)):
        if count >= min_template_occurrence:
            rulebase.add_rule(
                smarts=template, rule_hash=md5(template.encode()).hexdigest(), n_support=count
            )

    return rulebase, {rule.smarts: id for id, rule in rulebase.rules.items()}


def process_reactions(
    reactions: dict[DataFold, list[str]],
    min_template_occurrence: int = 0,
    wrap_smarts: bool = True,
    num_processes: int = cpu_count(),
) -> LabeledMolsAndTemplates:
    """Workflow for processing datasets of reactions.
       Performs
        1) template assignment on training set, and builds template library & index
        2) extracts products and assigns template labels to train, test, val datasets.
           Reactions not covered by the template library derived from the training data will receive a label of -1

    Args:
        reactions: Mapping from data fold (train, validation and test) to a list of reaction SMILES.
        min_template_occurrence: How often a template needs to appear in training data to receive a
            label. To reduce the number of false positives, it's often beneficial to increase this
            number.
        wrap_smarts: Whether to transform "reactants>conditions>products" into
            "(reactants)>conditions>(products)".
        num_processes: Number of parallel processes to use.

    Returns:
        Dataclass containing the labelled (mol, template) data, and the template library.
    """
    logger.info("Extracting templates from training set")
    processed_train_samples = extract_templates(
        reactions[DataFold.TRAIN], wrap_smarts=wrap_smarts, num_processes=num_processes
    )

    logger.info("Building the template library")
    rulebase, template_idx = rulebase_from_templates_list(
        templates=[cast(str, sample.template) for sample in processed_train_samples],
        min_template_occurrence=min_template_occurrence,
    )

    logger.info(f"Built template library with size {len(rulebase)}")

    samples = {}
    for fold in reactions:
        logger.info(f"Processing fold {fold}")

        if fold == DataFold.TRAIN:
            processed_samples = processed_train_samples
        else:
            processed_samples = extract_templates(
                reactions[fold], wrap_smarts=wrap_smarts, num_processes=num_processes
            )

        # Samples that use rare templates will be assigned a label of `-1`.
        samples[fold] = label(processed_samples, template_idx=template_idx)
        logger.info(f"{fold}: Original {len(reactions[fold])} samples")
        logger.info(f"{fold}: Processed {len(processed_samples)} samples")

    return LabeledMolsAndTemplates(samples=samples, rulebase=rulebase)


def label(
    samples: Iterable[TemplateReactionSample], template_idx: dict[str, int], rare_idx: int = -1
) -> Iterable[TemplateReactionSample]:
    """Attach template IDs to processed samples.

    Args:
        samples: Samples processed with `extract_templates`.
        template_idx: Template index i.e. mapping from SMARTS to ID.
        rare_idx: Label to assign when a template is not covered by the index.

    Returns:
        Iterable over data samples with template IDs attached.
    """
    for sample in samples:
        yield dataclasses.replace(
            sample,
            template=None,
            template_application_results=[
                TemplateApplicationResult(
                    template_id=template_idx.get(cast(str, sample.template), rare_idx)
                )
            ],
        )


def process_smiles(
    input_dir: str, output_dir: str, min_template_occurrence: int, num_processes: int
) -> None:
    logger.info(
        f"Params: {input_dir} {output_dir}, min template frequency: {min_template_occurrence}"
    )

    # Start by testing the `rdchiral` installation.
    rdchiral_result = template_extractor.extract_from_reaction(_convert("[C:1]>>[C:1]C", 0))
    if "reaction_smarts" not in rdchiral_result:
        raise RuntimeError("The rdchiral installation is not working.")
    elif "reaction_smarts_retro" not in rdchiral_result:
        logger.warning(
            "You seem to be using the original (Python) version of rdchiral. If data processing "
            "takes a long time, consider using the faster C++ version (rdchiral_cpp)."
        )

    output = process_reactions(
        reactions=load_raw_reactions_files(input_dir),
        min_template_occurrence=min_template_occurrence,
        num_processes=num_processes,
    )

    logger.info("Done processing, saving rulebase")
    output.rulebase.save_to_file(dir=output_dir)

    for fold, samples in output.samples.items():
        logger.info(f"Saving samples into fold {fold}")
        DiskReactionDataset.save_samples_to_file(data_dir=output_dir, fold=fold, samples=samples)  # type: ignore


def run_from_config(config: ExtractTemplatesConfig) -> None:
    set_random_seed(0)
    process_smiles(
        config.data_dir,
        config.output_dir,
        min_template_occurrence=config.min_freq,
        num_processes=config.num_processes,
    )


def main(argv: Optional[list[str]]) -> None:
    config: ExtractTemplatesConfig = cli_get_config(argv=argv, config_cls=ExtractTemplatesConfig)
    run_from_config(config)


if __name__ == "__main__":
    main(argv=None)



#!/usr/bin/env python
"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/translate/translator.py
Modifications:
1. Simplified the structure by removing the `build_translator` function, and merged the `Inference` class into the `Translator` class.
2. Refactored the `_run_encoder`, `_decode_and_generate`, and `_translate_batch_with_strategy` methods to align with the `retrochimera.models.smiles_transformer.SmilesTransformerModel` class.
3. Introduced the `customised_beam_search` attribute and corresponding logic to the `Translator` class, enabling optimized beam search for retrosynthesis prediction.
"""
from typing import Any, Optional

import torch
from torch.nn.functional import log_softmax

from retrochimera.data.smiles_tokenizer import Tokenizer
from retrochimera.models.smiles_transformer import SmilesTransformerModel
from retrochimera.opennmt.decode.beam_search import BeamSearch
from retrochimera.opennmt.decode.decoder_strategy import DecodeStrategy, set_random_seed


class Translator(object):
    """Translate a batch of sentences with a saved model and tokenizer.

    Args:
        model: Encoder-decoder model to use.
        tokenizer: Tokenizer.
        gpu: GPU device. Set to negative for no GPU.
        n_best: How many beams to wait for.
        min_length: See class .decode_strategy.DecodeStrategy.
        max_length: See class .decode_strategy.DecodeStrategy.
        ratio: See class .decode_strategy.DecodeStrategy.
        beam_size: Number of beams.
        block_ngram_repeat: See class .decode_strategy.DecodeStrategy.
        ignore_when_blocking: See class .decode_strategy.DecodeStrategy.
        replace_unk: Replace unknown token.
        ban_unk_token: Ban unknown token.
        tgt_prefix: Force the predictions begin with provided -tgt.
        data_type: Source data type.
        verbose: Print/log every translation.
        seed: Random seed.
    """

    def __init__(
        self,
        model: SmilesTransformerModel,
        tokenizer: Tokenizer,
        gpu: int = 0,
        n_best: int = 10,
        min_length: int = 0,
        max_length: int = 512,
        ratio: float = 0.0,
        beam_size: int = 10,
        block_ngram_repeat: int = 0,
        ignore_when_blocking: set = set(),
        replace_unk: bool = False,
        ban_unk_token: bool = False,
        tgt_prefix: bool = False,
        data_type: str = "text",
        verbose: bool = False,
        seed: int = 1,
        customised_beam_search: bool = True,
    ):
        self.model = model
        self.tokenizer = tokenizer

        self._tgt_eos_idx = self.tokenizer.end_token_id
        self._tgt_pad_idx = self.tokenizer.pad_token_id
        self._tgt_bos_idx = self.tokenizer.begin_token_id
        self._tgt_unk_idx = self.tokenizer.unk_token_id

        self._tgt_vocab = tokenizer.vocab_dict  # token -> id
        self._tgt_vocab_len = len(self._tgt_vocab)

        self.gpu = gpu
        self.use_cuda = gpu > -1
        self.device = torch.device("cuda", self.gpu) if self.use_cuda else torch.device("cpu")

        self.n_best = n_best
        self.min_length = min_length
        self.max_length = max_length

        self.ratio = ratio
        self.beam_size = beam_size

        self.block_ngram_repeat = block_ngram_repeat
        self.ignore_when_blocking = ignore_when_blocking
        self._exclusion_idxs = {self._tgt_vocab[t] for t in self.ignore_when_blocking}
        self.replace_unk = replace_unk
        if self.replace_unk:
            raise ValueError("replace_unk requires an attentional decoder.")
        self.ban_unk_token = ban_unk_token

        self.tgt_prefix = tgt_prefix
        self.data_type = data_type
        self.verbose = verbose
        self.customised_beam_search = customised_beam_search

        set_random_seed(seed, self.use_cuda)

    def translate_batch(self, batch: dict[str, Any], attn_debug: bool = False) -> dict[str, Any]:
        """Translate a batch of sentences.

        Args:
            batch: dict, keys:
            - src: Tuple(src, src_lengths)
                - src (torch.Tensor): shape of src: (padded_src_len, batch_size, 1)
                - src_lengths (torch.Tensor): shape of src_lengths: (batch_size,)
            - batch_size: int
        """
        with torch.no_grad():
            # TODO: support these blacklisted features
            decode_strategy = BeamSearch(
                pad=self._tgt_pad_idx,
                bos=self._tgt_bos_idx,
                eos=self._tgt_eos_idx,
                unk=self._tgt_unk_idx,
                batch_size=batch["batch_size"],
                beam_size=self.beam_size,
                n_best=self.n_best,
                min_length=self.min_length,
                max_length=self.max_length,
                block_ngram_repeat=self.block_ngram_repeat,
                exclusion_tokens=self._exclusion_idxs,
                ratio=self.ratio,
                ban_unk_token=self.ban_unk_token,
                return_attention=attn_debug or self.replace_unk,
                customised_beam_search=self.customised_beam_search,
            )

            return self._translate_batch_with_strategy(batch, decode_strategy)

    def _translate_batch_with_strategy(
        self, batch: dict[str, Any], decode_strategy: DecodeStrategy
    ) -> dict[str, Any]:
        """Translate a batch of sentences step by step using cache.

        Args:
            batch (dict), keys:
            - src: Tuple(src, src_lengths)
                - src (torch.Tensor): shape of src: (padded_src_len, batch_size, 1)
                - src_lengths (torch.Tensor): shape of src_lengths: (batch_size,)
            - batch_size: int

            decode_strategy (DecodeStrategy): A decode strategy to use for
                generate translation step by step.

        Returns:
            results (dict): The translation results.
        """
        # (0) Prep the components of the search.
        use_src_map = False
        parallel_paths = decode_strategy.parallel_paths  # beam_size
        batch_size = batch["batch_size"]

        # (1) Run the encoder on the src.
        src, encoder_embedding, memory_bank, src_lengths = self._run_encoder(batch)
        # shape of src: (padded_src_len, batch_size, 1)
        # shape of encoder_embedding: (padded_src_len, batch_size, hidden_dim)
        # shape of memory_bank: (padded_src_len, batch_size, hidden_dim)
        # shape of src_lengths: (batch_size,)

        # self.model.decoder.init_state(src, memory_bank, encoder_embedding)
        self.model.decoder.init_state(src, None, None)

        gold_score = [0] * batch_size

        # (2) prep decode_strategy. Possibly repeat src objects.
        src_map = batch["src_map"] if use_src_map else None  # None

        target_prefix = batch["tgt"] if self.tgt_prefix else None  # None

        (fn_map_state, memory_bank, memory_lengths, src_map,) = decode_strategy.initialize(
            memory_bank, src_lengths, src_map, target_prefix=target_prefix
        )
        # [returns]
        # shape of memory_bank: (padded_src_len, batch_size * beam_size, hidden_dim)
        # shape of memory_lengths: (batch_size * beam_size,)

        if fn_map_state is not None:
            self.model.decoder.map_state(fn_map_state, only_map_src=True)

        # (3) Begin decoding step by step:
        for step in range(decode_strategy.max_length):
            decoder_input = decode_strategy.current_predictions.view(
                1, -1, 1
            )  # type: ignore[override]
            # shape of decoder_input: (1, self.batch_size * self.parallel_paths, 1)

            log_probs, attn = self._decode_and_generate(
                decoder_input,
                memory_bank,
                batch,
                memory_lengths=memory_lengths,
                src_map=src_map,
                step=step,
                batch_offset=decode_strategy.batch_offset,  # type: ignore
            )  # (batch_size * beam_size, vocab)

            if self.customised_beam_search:
                # modify the log_probs for finished sentences.
                _, vocab_size = tuple(log_probs.shape)
                bad_token_log_prob = -1e5

                complete_seq_log_prob = (torch.ones((1, vocab_size)) * bad_token_log_prob).to(
                    log_probs.device
                )  # shape: (1, vocab_size)
                complete_seq_log_prob[:, self._tgt_eos_idx] = 0.0  # shape: (1, vocab_size)

                # Use this vector in the output for sequences which are complete.
                is_end_token = (
                    decoder_input.squeeze() == self._tgt_eos_idx
                )  # shape: (batch_size * beam_size,)
                log_prob_mask = is_end_token.unsqueeze(1)  # shape: (batch_size * beam_size, 1)
                log_probs = (log_prob_mask * complete_seq_log_prob) + (
                    ~log_prob_mask * log_probs
                )  # shape: (batch_size * beam_size, vocab_size)

                assert log_probs.dim() == 2, f"Expected 2D tensor, got {log_probs.dim()}"
                assert log_probs.size(0) == decoder_input.size(
                    1
                ), f"Expected {batch_size * parallel_paths}, got {log_probs.size(0)}"
                assert (
                    log_probs.size(1) == vocab_size
                ), f"Expected {vocab_size}, got {log_probs.size(1)}"

            decode_strategy.advance(log_probs, attn)
            any_finished = decode_strategy.is_finished.any()
            if any_finished:
                decode_strategy.update_finished()
                if decode_strategy.done:
                    break

            select_indices = decode_strategy.select_indices  # type: ignore

            if any_finished:
                # Reorder states.
                if isinstance(memory_bank, tuple):
                    memory_bank = tuple(x.index_select(1, select_indices) for x in memory_bank)
                else:
                    memory_bank = memory_bank.index_select(1, select_indices)

                memory_lengths = memory_lengths.index_select(0, select_indices)

                if src_map is not None:
                    src_map = src_map.index_select(1, select_indices)

            if parallel_paths > 1 or any_finished:
                self.model.decoder.map_state(
                    lambda state, dim: state.index_select(dim, select_indices)
                )

        return self.report_results(
            gold_score,
            batch,
            batch_size,
            decode_strategy,
        )

    def _run_encoder(self, batch: dict[str, Any]) -> tuple:
        src, src_lengths = (
            batch["src"] if isinstance(batch["src"], tuple) else (batch["src"], None)
        )  # src: (padded_src_len, batch_size, 1), src_lengths: (batch_size,)

        encoder_embedding = self.model.construct_input(
            src.transpose(0, 1).contiguous(), is_encoder=True
        )  # shape: (batch_size, padded_src_len, hidden_dim)
        encoder_padding_mask = (
            src.transpose(0, 1).contiguous().squeeze(2).eq(self._tgt_pad_idx)
        )  # shape: (batch_size, padded_src_len)

        memory_bank = (
            self.model.encoder.forward(
                src=encoder_embedding, src_key_padding_mask=encoder_padding_mask
            )
            .transpose(0, 1)
            .contiguous()
        )  # shape: (padded_src_len, batch_size, hidden_dim)

        if src_lengths is None:
            assert not isinstance(
                memory_bank, tuple
            ), "Ensemble decoding only supported for text data"
            src_lengths = (
                torch.Tensor(batch["batch_size"])
                .type_as(memory_bank)
                .long()
                .fill_(memory_bank.size(0))
            )  # shape: (batch_size,)
        # shape of src: (padded_src_len, batch_size, 1)
        # shape of encoder_embedding.transpose(0,1).contiguous(): (padded_src_len, batch_size, hidden_dim)
        # shape of memory_bank: (padded_src_len, batch_size, hidden_dim)
        # shape of src_lengths: (batch_size,)
        return src, encoder_embedding.transpose(0, 1).contiguous(), memory_bank, src_lengths

    def _decode_and_generate(
        self,
        decoder_in: torch.Tensor,
        memory_bank: torch.Tensor,
        batch: dict[str, Any],
        memory_lengths: torch.Tensor,
        src_map=None,
        step: Optional[int] = None,
        batch_offset: torch.LongTensor = None,
    ):
        """Decode and generate one step.

        Args:
            decoder_in (torch.Tensor): shape: (1, batch_size * beam_size, 1), due to kv_cache mechanism
            memory_bank (torch.Tensor): shape: (padded_src_len, batch_size * beam_size, hidden_dim)
            batch (dict), keys:
            - src: Tuple(src, src_lengths)
                - src (torch.Tensor): shape of src: (padded_src_len, batch_size, 1)
                - src_lengths (torch.Tensor): shape of src_lengths: (batch_size,)
            - batch_size: int
            memory_lengths (torch.Tensor): shape: (batch_size * beam_size,)
            src_map (torch.Tensor): None
            step (int): current step
            batch_offset (int): batch offset

        Returns:
            log_probs (torch.Tensor): shape: (batch_size * beam_size, vocab)
            attn (torch.Tensor): shape: (batch_size * beam_size, tgt_len, src_len)
        """
        # dec_out, dec_attn = self.model.decoder(
        #     decoder_in, memory_bank, memory_lengths=memory_lengths, step=step
        # )

        # # Generator forward.
        # if "std" in dec_attn:
        #     attn = dec_attn["std"]
        # else:
        #     attn = None
        # log_probs = self.model.generator(dec_out.squeeze(0))

        # return log_probs, attn

        # Decoder forward, takes [batch, tgt_len, nfeats] as input
        # and [batch, src_len, hidden] as enc_out
        # in case of inference tgt_len = 1, batch = beam times batch_size
        # in case of Gold Scoring tgt_len = actual length, batch = 1 batch

        decoder_embedding = self.model.construct_input(
            decoder_in.transpose(0, 1).contiguous(), step=step, is_encoder=False
        )  # shape: (batch_size * beam_size, 1, hidden_dim)
        assert decoder_embedding.dim() == 3, f"Expected 3D tensor, got {decoder_embedding.dim()}"

        decoder_padding_mask = decoder_in.eq(self._tgt_pad_idx).squeeze(
            2
        )  # shape: (1, batch_size * beam_size)
        memory_padding_mask = torch.arange(
            0, max(memory_lengths), device=memory_bank.device
        ) >= memory_lengths.unsqueeze(
            1
        )  # shape: (batch_size * beam_size, padded_src_len)

        dec_out, dec_attn = self.model.decoder.forward(
            tgt=decoder_embedding,  # shape: (batch_size * beam_size, 1, hidden_dim)
            tgt_key_padding_mask=decoder_padding_mask.transpose(
                0, 1
            ).contiguous(),  # shape: (batch_size * beam_size, 1)
            enc_out=memory_bank.transpose(
                0, 1
            ).contiguous(),  # shape: (batch_size * beam_size, padded_src_len, hidden_dim)
            enc_key_padding_mask=memory_padding_mask,  # shape: (batch_size * beam_size, padded_src_len)
            step=step,
        )  # shape: (batch_size, tgt_len-1, hidden_dim)

        # Generator forward.
        if "std" in dec_attn:
            attn = dec_attn["std"]
        else:
            attn = None
        # scores = self.model.generator(dec_out.squeeze(1))
        scores = self.model.token_fc(
            dec_out.squeeze(1)
        )  # shape: (batch_size * beam_size, vocab_size)
        log_probs = log_softmax(scores, dim=-1)  # shape: (batch_size * beam_size, vocab_size)
        # returns [(batch_size x beam_size) , vocab ] when 1 step
        # or [batch_size, tgt_len, vocab ] when full sentence

        return log_probs, attn

    def report_results(
        self,
        gold_score,
        batch,
        batch_size,
        decode_strategy,
    ):
        results = {
            "predictions": None,
            "scores": None,
            "attention": None,
            "batch": batch,
            "gold_score": gold_score,
        }

        results["scores"] = decode_strategy.scores
        results["predictions"] = decode_strategy.predictions
        results["attention"] = decode_strategy.attention
        results["alignment"] = [[] for _ in range(batch_size)]
        return results



import pydoc
import re
from dataclasses import fields, is_dataclass
from typing import Any, Union, get_type_hints

import h5py
import numpy as np
import scipy
from rdkit import RDLogger


def save_into_h5py(group: h5py.Group, key: str, data: Any) -> None:
    """Save a given object into a `h5py.Group`.

    This function only works for the processed data sample dataclasses and their subcomponents.
    """
    cls = type(data)
    cls_name = f"{cls.__module__}.{cls.__name__}"

    if is_dataclass(data):
        data_dict = {field.name: getattr(data, field.name) for field in fields(data)}
    elif isinstance(data, scipy.sparse.dok_matrix):
        assert data.dtype == np.int32

        data_dict = {
            "shape": np.asarray(data.shape),
            "keys": np.asarray(list(data.keys())),
            "values": np.asarray(list(data.values())),
        }
    else:
        dataset = group.create_dataset(key, data=data)
        dataset.attrs["type"] = cls_name
        return

    subgroup = group.create_group(key)
    subgroup.attrs["type"] = cls_name

    for subkey, value in data_dict.items():
        if value is not None:
            save_into_h5py(subgroup, key=subkey, data=value)


def load_from_h5py(entry: Union[h5py.Dataset, h5py.Group]) -> Any:
    """Parse a given class from a `h5py` entry.

    This function only works for the processed data sample dataclasses and their subcomponents.
    """
    cls_name = entry.attrs["type"]
    cls = pydoc.locate(cls_name)

    if is_dataclass(cls):
        assert isinstance(entry, h5py.Group)

        type_hints = get_type_hints(cls)
        return cls(
            **{
                key: (load_from_h5py(entry[key]) if key in entry else None)
                for key in type_hints.keys()
            }
        )  # type: ignore
    elif cls is scipy.sparse.dok_matrix:
        assert isinstance(entry, h5py.Group)

        matrix = scipy.sparse.dok_matrix(tuple(entry["shape"]), dtype="i")
        for (key_row, key_col), value in zip(entry["keys"], entry["values"]):
            matrix[key_row, key_col] = value

        return matrix
    elif cls in (int, bool):
        return np.asarray(entry).item()
    elif cls is list:
        return list(entry)
    else:
        assert isinstance(entry, h5py.Dataset)
        return np.asarray(entry)


def lookup_by_name(module, name):
    return module.__dict__[name]


def silence_rdkit_warnings() -> None:
    RDLogger.DisableLog("rdApp.*")


def convert_camel_to_snake(name: str) -> str:
    """Convert a CamelCase name to snake_case.

    Consecutive capital letters (e.g. "GNN") are treated as a single word.

    >>> convert_camel_to_snake('TemplateLocalization')
    'template_localization'

    >>> convert_camel_to_snake('TemplateClassificationGNN')
    'template_classification_gnn'

    >>> convert_camel_to_snake('MEGAN')
    'megan'
    """
    return re.sub("(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])", "_", name).lower()



import random


def get_product_roots(product_atom_ids: list[int], num_augmentations: int) -> list[int]:
    product_roots = [-1]

    if len(product_atom_ids) < num_augmentations:
        product_roots.extend(product_atom_ids)
        product_roots.extend(
            random.choices(product_roots, k=num_augmentations - len(product_roots))
        )
    else:
        product_roots.extend(random.sample(product_atom_ids, num_augmentations - 1))

    assert len(product_roots) == num_augmentations
    return product_roots



import random
import tempfile

import pytest
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample
from syntheseus.reaction_prediction.utils.misc import set_random_seed

from retrochimera.chem.rule_application_server import RuleApplicationServer
from retrochimera.chem.rules import get_products
from retrochimera.data.dataset import DataFold, ReactionDataset


@pytest.mark.parametrize("num_inputs", [1, 16])
@pytest.mark.parametrize("num_processes", list(range(1, 6)))
def test_apply_rules(reaction_dataset: ReactionDataset, num_inputs: int, num_processes: int):
    # Set a deterministic seed, but make it vary between tests for better coverage.
    set_random_seed(hash((num_inputs, num_processes)) % (2**30))

    with tempfile.TemporaryDirectory() as temp_dir:
        # Save the rulebase and let the processes load it. The file will be deleted when we exit the
        # context, but once the server constructor returns the file would only be needed if rule
        # application processes need to be restarted (which shouldn't happen here).
        reaction_dataset.rulebase.save_to_file(dir=temp_dir)
        server = RuleApplicationServer(rulebase_dir=temp_dir, num_processes=num_processes)

    # Take a few random products from the dataset.
    samples: list[ReactionSample] = random.sample(
        list(reaction_dataset[DataFold.TRAIN]), num_inputs
    )
    inputs = [list(sample.products)[0] for sample in samples]

    # Pair the inputs with a random number of randomly chosen rule IDs.
    rule_ids_to_apply = [
        [random.randint(0, 3) for _ in range(random.randint(2, 6))] for _ in inputs
    ]

    # Apply the rules in parallel.
    results_parallel = server.apply_rules(inputs=inputs, rule_ids_to_apply=rule_ids_to_apply)

    assert len(results_parallel) == len(inputs)

    for input, rule_ids, results in zip(inputs, rule_ids_to_apply, results_parallel):
        # Apply the same rules naively and check that the outcome is the same.
        results_naive = []
        for rule_id in rule_ids:
            results_naive.extend(get_products(input, rule=reaction_dataset.rulebase[rule_id].rxn))

        assert sum(results, []) == results_naive



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, Iterable, Optional, Union

import torch
from torch.nn import functional as f

from retrochimera.chem.rules import RuleBase
from retrochimera.data.preprocessing.template_classification import (
    ProcessedSample,
    preprocess_samples,
)
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.encoders.base import BatchType, get_encoder_by_name
from retrochimera.layers.mlp import MLP
from retrochimera.models.lightning import AbstractModel, log_classification_metrics


@dataclass
class Batch(Generic[BatchType]):
    inputs: BatchType
    targets: Any


class MCCModel(AbstractModel[TemplateReactionSample, ProcessedSample, Batch]):
    """Model performing template classification on top of an arbitrary encoder.

    This model simply takes the molecule-level output from a backbone `Encoder`, and passes that
    through an MLP to compute template probabilities.
    """

    def __init__(
        self,
        encoder_class: str,
        encoder_kwargs: dict[str, Any],
        hidden_dim: int,
        n_classes: int,
        n_hidden_layers: int = 1,
        dropout: float = 0.2,
        label_smoothing: float = 0.0,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.hidden_dim = hidden_dim
        self.n_classes = n_classes
        self.n_hidden_layers = n_hidden_layers
        self.encoder_class = encoder_class
        self.encoder_kwargs = encoder_kwargs
        self.dropout = dropout
        self._label_smoothing = label_smoothing
        self.rulebase_dir: Optional[Union[str, Path]] = None

        self.rebuild()

    def set_rulebase(self, rulebase: RuleBase, rulebase_dir: Union[str, Path]) -> None:
        self.rulebase_dir = rulebase_dir

    def rebuild(self):
        self.encoder = get_encoder_by_name(self.encoder_class)(**self.encoder_kwargs).to(
            self.device
        )
        self.input_dim = self.encoder.mol_out_channels
        self.mlp = MLP(
            self.input_dim,
            hidden_dim=self.hidden_dim,
            out_dim=self.n_classes,
            n_layers=self.n_hidden_layers,
            dropout=self.dropout,
        ).to(self.device)

    def forward(self, x):
        return self.mlp(x)

    def ttv_step(self, batch: Batch, step_name: str):
        batch_encoded = self.encoder(batch.inputs)
        batch_size = batch_encoded.mol_outputs.shape[0]

        y_hat = self.forward(batch_encoded.mol_outputs)
        loss = f.cross_entropy(y_hat, batch.targets, label_smoothing=self._label_smoothing)

        self.log(f"{step_name}_loss", loss, batch_size=batch_size)
        log_classification_metrics(
            log_fn=self.log,
            step_name=step_name,
            batch_preds=y_hat,
            batch_targets=batch.targets,
            batch_size=batch_size,
        )

        return loss

    def preprocess(
        self, samples: Iterable[TemplateReactionSample], num_processes: int = 0
    ) -> Iterable[ProcessedSample]:
        assert self.rulebase_dir is not None
        yield from preprocess_samples(
            samples=samples,
            rulebase_dir=self.rulebase_dir,
            encoder=self.encoder,
            num_processes=num_processes,
        )

    def collate(self, samples: list[ProcessedSample]) -> Batch:
        return Batch(
            inputs=self.encoder.collate([sample.input for sample in samples]),
            targets=torch.as_tensor([sample.target for sample in samples], dtype=torch.long),
        )



from rdkit import Chem

from retrochimera.chem.fixing import fix_mol, reverse_rxn


def test_fix_mol_possible() -> None:
    smi = "c1cc2cc(ccc2n1)-c1ncno1"
    out = fix_mol(Chem.MolFromSmiles(smi, sanitize=False))
    tgt = Chem.CanonSmiles("N1C=CC2=CC(=CC=C12)C1=NC=NO1")
    assert out == tgt


def test_fix_mol_impossible() -> None:
    # First of the SMILES below is already rejected by `adjust_aromatic_Ns`, second one is processed
    # successfully but at the end still cannot be sanitized.
    broken_smiles = ["nnn", "COc1ccccc1Oc1nc(Nc2cc(C)[nH](C3CCCCO3)n2)cc2ccccc12"]

    for smiles in broken_smiles:
        assert fix_mol(Chem.MolFromSmiles(smiles, sanitize=False)) is None


def test_reverse() -> None:
    assert reverse_rxn("A>B>C.D") == "C.D>B>A"
    assert reverse_rxn("A>>C1OC1") == "C1OC1>>A"



"""Script for evaluating a model on a given dataset.

Each of the model types can be loaded from a *single directory*, possibly containing several files
(e.g. checkpoint, config, etc). See individual model wrappers for the model directory formats.

Example invocation:
    python ./retrochimera/cli/eval.py \
        data_dir=[DATA_DIR] \
        model_class=TemplateLocalization \
        model_dir=[MODEL_DIR]
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from omegaconf import MISSING
from syntheseus.cli import eval_single_step as eval_syntheseus
from syntheseus.reaction_prediction.inference.config import ModelConfig
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config

from retrochimera.inference.retrochimera import RetroChimeraModel
from retrochimera.inference.smiles_transformer import SmilesTransformerModel
from retrochimera.inference.smiles_transformer_forward import SmilesTransformerForwardModel
from retrochimera.inference.template_classification import TemplateClassificationModel
from retrochimera.inference.template_localization import TemplateLocalizationModel


class BackwardModelClass(Enum):
    RetroChimera = RetroChimeraModel
    SmilesTransformer = SmilesTransformerModel
    SmilesTransformerForward = SmilesTransformerForwardModel
    TemplateClassification = TemplateClassificationModel
    TemplateLocalization = TemplateLocalizationModel


@dataclass
class BackwardModelConfig(ModelConfig):
    """Config for loading one of the supported backward models."""

    model_class: BackwardModelClass = MISSING


@dataclass
class EvalConfig(BackwardModelConfig, eval_syntheseus.BaseEvalConfig):
    """Config for running evaluation on a given dataset."""

    pass


def main(argv: Optional[list[str]]) -> None:
    config: EvalConfig = cli_get_config(argv=argv, config_cls=EvalConfig)
    eval_syntheseus.run_from_config(config, extra_steps=[])  # type: ignore


if __name__ == "__main__":
    main(argv=None)



"""Integration tests for the reaction prediction pipeline.

These tests should be executed **in this order**, and use the shared temporary directory to avoid
duplicating work (e.g. data preprocessing). One downside of separating interconnected bits of
testing is that if one test fails, subsequent ones will likely fail as well; however, we also get
cleaner error reporting than if everything was one big test.
"""

import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Any, Optional

import pytest
from syntheseus import Bag, Molecule

from retrochimera.utils.misc import convert_camel_to_snake

NUM_SAMPLES = 30  # For truncating training data.


def run_with_python(path: str, args: list[str]) -> subprocess.CompletedProcess:
    return subprocess.run(
        [sys.executable, "-m", "coverage", "run", "--source", "./retrochimera/", "-p", path] + args,
        capture_output=True,
        check=True,
    )


def get_num_lines(path: Path) -> int:
    with open(path, "r") as f:
        return len(f.readlines())


def get_num_files(dir: Path) -> int:
    return len([f for f in os.listdir(dir)])


def test_deduplicate_dataset(deduplicate_output_path: Path) -> None:
    """Test deduplication of reactions in the dataset."""
    raw_data_path = Path("./tests/testdata/uspto_reactions.smi")
    assert get_num_lines(raw_data_path) == 630

    run_with_python(
        path="./retrochimera/cli/deduplicate_dataset.py",
        args=[f"input_path={raw_data_path}", f"output_path={deduplicate_output_path}"],
    )

    # Test data contains 5 artificially inserted duplicates.
    assert get_num_lines(deduplicate_output_path) == 625


def test_split_dataset(split_output_dir: Path, deduplicate_output_path: Path) -> None:
    """Test splitting the dataset into training, validation, and test sets."""
    run_with_python(
        path="./retrochimera/cli/split_dataset.py",
        args=[
            f"input_path={deduplicate_output_path}",
            f"output_dir={split_output_dir}",
            "val_frac=0.05",
            "test_frac=0.1",
            "max_reactants_num=4",
            "min_product_atoms=3",
            "max_product_atoms=60",
        ],
    )

    # Test data contains real reactions from USPTO-50K plus 5 extra ones that get filtered out:
    # - [C:1].C.C.C.C>>[C:1]CCCCC    too many reactants
    # - [C:1]-[C:1]>>[C:1]CCCC       double mapped left-hand side
    # - [C:1].[O:2]>>[C:1][O:2]      product too small
    # - [C:1].CCC...>>[C:1]CCC...    product too large
    # - [C:1]CCCC.O>>[C:1]CCCC       product among the reactants

    assert get_num_lines(split_output_dir / "pista_train.smi") == 527
    assert get_num_lines(split_output_dir / "pista_val.smi") == 31
    assert get_num_lines(split_output_dir / "pista_test.smi") == 62


def test_extract_templates(split_output_dir: Path, processed_output_dir: Path) -> None:
    """Test extraction of templates from the dataset."""
    run_with_python(
        path="./retrochimera/cli/extract_templates.py",
        args=[f"data_dir={split_output_dir}", f"output_dir={processed_output_dir}"],
    )

    EXPECTED_SIZE_TRAIN = 527
    EXPECTED_SIZE_VAL = 31
    EXPECTED_SIZE_TEST = 62

    assert get_num_lines(processed_output_dir / "template_lib.json") == 22
    assert get_num_lines(processed_output_dir / "train.jsonl") == EXPECTED_SIZE_TRAIN
    assert get_num_lines(processed_output_dir / "val.jsonl") == EXPECTED_SIZE_VAL
    assert get_num_lines(processed_output_dir / "test.jsonl") == EXPECTED_SIZE_TEST


def test_augment_rsmiles(processed_output_dir: Path, augmented_output_dir: Path) -> None:
    """Test R-SMILES data augmentation (without using the augmented data for training)."""
    NUM_AUGMENTATIONS = 10

    run_with_python(
        path="./retrochimera/cli/augment_rsmiles.py",
        args=[
            f"data_dir={processed_output_dir}",
            f"augmented_data_dir={augmented_output_dir}",
            f"augmentation={NUM_AUGMENTATIONS}",
        ],
    )

    for (path_input, path_output) in [
        (processed_output_dir / "train.jsonl", augmented_output_dir / "train" / "train.jsonl"),
        (processed_output_dir / "val.jsonl", augmented_output_dir / "val" / "val.jsonl"),
    ]:
        with open(path_input, "r") as f_in, open(path_output, "r") as f_out:
            rxns_input = [json.loads(line) for line in f_in.readlines()]
            rxns_output = [json.loads(line) for line in f_out.readlines()]

        assert len(rxns_output) == NUM_AUGMENTATIONS * len(rxns_input)

        for idx, rxn in enumerate(rxns_input):
            rxns_augmented = rxns_output[idx * NUM_AUGMENTATIONS : (idx + 1) * NUM_AUGMENTATIONS]

            # Check augmentation produced varying outputs.
            for key in ["reactants", "products"]:
                values = [tuple(d["smiles"] for d in r[key]) for r in rxns_augmented]
                assert len(set(values)) >= NUM_AUGMENTATIONS // 2

                # Check that all outputs match the original when canonicalized.
                for value in values:
                    assert Bag(Molecule(smiles) for smiles in value) == Bag(
                        [Molecule(d["smiles"]) for d in rxn[key]]
                    )

            # Check mapped reaction SMILES is the same across augmentations.
            for rxn_augmented in rxns_augmented:
                assert rxn_augmented["mapped_reaction_smiles"] == rxn["mapped_reaction_smiles"]

    # Use a small sample of training data as all folds to let the model easily overfit. We do this
    # *after* testing augmentation to still cover the case of augmenting varied folds.
    processed_train_set_path = processed_output_dir / "train.jsonl"

    with open(processed_train_set_path, "rt") as f:
        lines = f.readlines()
        lines = lines[:NUM_SAMPLES]

    with open(processed_train_set_path, "wt") as f:
        f.writelines(lines)

    for fold in ["val", "test"]:
        shutil.copy(processed_train_set_path, processed_output_dir / f"{fold}.jsonl")


@pytest.mark.parametrize(
    ("testcase", "extra_model_kwargs"),
    [
        ("TemplateClassificationChem", {}),
        ("TemplateClassificationGNN", {}),
        ("TemplateLocalization[softmax]", {}),
        ("TemplateLocalization[sigmoid]", {"classification_loss_type": "sigmoid"}),
        ("SmilesTransformer", {}),
    ],
)
def test_preprocess_and_train(
    testcase: str,
    extra_model_kwargs: dict[str, Any],
    tmp_path: Path,
    processed_output_dir: Path,
) -> None:
    """Test the end-to-end reaction prediction pipeline for various model classes."""
    model_class = testcase.split("[")[0]

    # Prepare kwargs that define a small GNN model to reuse across several model classes.
    small_gnn_kwargs = [
        f"encoder.{k}"
        for k in [
            "atom_out_channels=16",
            "mol_out_channels=512",
            "atom_categorical_features_channels=null",
            "aggregation_num_heads=1",
            "aggregation_dropout=0.0",
            "gnn_kwargs.hidden_channels=64",
            "gnn_kwargs.num_layers=1",
            "gnn_kwargs.dropout=0.0",
        ]
    ]

    # Prepare kwargs for all models (mostly making them smaller).
    model_class_to_kwargs = {
        "TemplateClassificationChem": ["encoder.fingerprint_dim=509"],
        "TemplateClassificationGNN": small_gnn_kwargs,
        "TemplateLocalization": sum([[f"input_{k}", f"rewrite_{k}"] for k in small_gnn_kwargs], [])
        + [
            "classification_label_smoothing=0.1",
            "classification_space_dim=256",
            "free_rewrite_embedding_dim=0",
            "num_negative_rewrites_in_localization=0",
            "negative_to_positive_targets_ratio=1.5",
            "rewrite_encoder_num_epochs=null",
            "training.batch_size=8",
        ],
        "SmilesTransformer": [
            "hidden_dim=32",
            "feedforward_dim=32",
            "n_layers=1",
            "n_heads=2",
            "schedule=constant",
            "warm_up_steps=5",
            "dropout=0.0",
        ],
    }

    train_config_key = f"{convert_camel_to_snake(model_class)}_config"

    featurized_output_dir = tmp_path / f"featurized_{testcase}"
    checkpoint_dir = tmp_path / f"checkpoint_{testcase}"
    log_dir = tmp_path / f"logs_{testcase}"
    eval_results_dir = tmp_path / f"eval_results_{testcase}"

    if model_class.startswith("TemplateClassification"):
        eval_model_class = "TemplateClassification"
    else:
        eval_model_class = model_class

    # Assemble model kwargs which have to be consistent between preprocessing and training.
    model_kwargs = [f"model_class={model_class}"] + [
        f"{train_config_key}.{k}" for k in model_class_to_kwargs[model_class]
    ]

    # Append any extra, test-specific kwargs.
    model_kwargs.extend([f"{train_config_key}.{k}={v}" for k, v in extra_model_kwargs.items()])

    preprocess_args = model_kwargs + [
        f"data_dir={processed_output_dir}",
        f"processed_data_dir={featurized_output_dir}",
        "rulebase_min_rule_support=2",
        "num_processes_preprocessing=1",
    ]

    n_epochs = 450 if model_class == "SmilesTransformer" else 200
    train_args = model_kwargs + [
        f"processed_data_dir={featurized_output_dir}",
        f"checkpoint_dir={checkpoint_dir}",
        f"log_dir={log_dir}",
        f"{train_config_key}.training.n_epochs={n_epochs}",
        f"{train_config_key}.training.learning_rate=0.001",
        f"{train_config_key}.training.learning_rate_decay_step_size={n_epochs // 2}",
        f"{train_config_key}.training.check_val_every_n_epoch=50",
        f"{train_config_key}.training.num_checkpoints_for_averaging=2",
        f"{train_config_key}.training.accelerator=cpu",
        f"{train_config_key}.training.gradient_clip_val=50.0",
        f"{train_config_key}.training.accumulate_grad_batches=1",
    ]

    extra_args = []
    if model_class == "SmilesTransformer":
        featurized_output_dir.mkdir()
        vocab_output_path = featurized_output_dir / "vocab.txt"

        run_with_python(
            path="./retrochimera/cli/build_tokenizer.py",
            args=[f"data_dir={processed_output_dir}", f"output_vocab_path={vocab_output_path}"],
        )

        assert vocab_output_path.exists()

        vocab_arg = f"{train_config_key}.vocab_path={vocab_output_path}"
        train_args += [vocab_arg]
        preprocess_args += [vocab_arg]

        # We did not apply augmentation to the training data, so turn it off for eval.
        extra_args += ["model_kwargs={'augmentation_size': 1}"]

    run_with_python(path="./retrochimera/cli/preprocess.py", args=preprocess_args)

    # Test data is composed of reactions for three most common templates in USPTO-50K, plus some
    # extra reactions using uncommon templates; the latter should get filtered out.
    assert get_num_lines(featurized_output_dir / "template_lib.json") == 3
    assert (featurized_output_dir / "data.h5").exists()

    run_with_python(path="./retrochimera/cli/train.py", args=train_args)

    # As many best checkpoints as specified via `num_checkpoints_for_averaging`.
    assert get_num_files(checkpoint_dir / "best") == 2
    assert get_num_files(checkpoint_dir / "last") == 1
    assert (checkpoint_dir / "combined.ckpt").exists()

    check_model_eval(
        eval_model_class=eval_model_class,
        model_dir=checkpoint_dir,
        data_dir=processed_output_dir,
        eval_results_dir=eval_results_dir,
        extra_args=extra_args,
    )


def check_model_eval(
    eval_model_class: str,
    model_dir: Path,
    data_dir: Path,
    eval_results_dir: Path,
    extra_args: Optional[list[str]] = None,
) -> None:
    """Run model evaluation and check the results."""
    eval_args = [
        f"model_class={eval_model_class}",
        f"model_dir={model_dir}",
        f"data_dir={data_dir}",
        f"results_dir={eval_results_dir}",
        "num_gpus=0",
    ]

    run_with_python(path="./retrochimera/cli/eval.py", args=eval_args + (extra_args or []))

    assert get_num_files(eval_results_dir) == 1
    [eval_results_path] = os.listdir(eval_results_dir)

    with open(eval_results_dir / eval_results_path, "r") as f:
        eval_results = json.load(f)

    assert eval_results["eval_args"]["model_class"] == f"{eval_model_class}Model"
    assert eval_results["eval_args"]["fold"] == "TEST"
    assert eval_results["num_samples"] == NUM_SAMPLES

    # In this example, all template-based are capped at ~90% due to a small fraction of examples
    # using out-of-library templates. They should however be able to overfit the rest.
    assert eval_results["top_k"][0] > 0.8


def test_optimize_ensembles(tmp_path: Path, processed_output_dir: Path) -> None:
    """Test the creation of a model ensemble."""
    # Copy over model results for ensembling.
    ensembling_dir = tmp_path / "ensembling"

    model_classes = ["TemplateLocalization", "SmilesTransformer"]
    testcases = ["TemplateLocalization[softmax]", "SmilesTransformer"]

    # Ensembling is designed to be robust to longer names (e.g. `template_localization_v1`), which
    # should then get shortened (to e.g. `template_localization`) when saving ensemble checkpoint.
    model_class_to_long_lowercase_name = {
        model_class: convert_camel_to_snake(model_class) + f"_v{i}"
        for i, model_class in enumerate(model_classes)
    }

    for model_class, testcase in zip(model_classes, testcases):
        # Ensembling needs both validation and test outputs. We only ran the latter, but since the
        # folds are the same, we just copy the results twice and adjust the fold key as needed.

        for fold in ["validation", "test"]:
            target_dir = (
                ensembling_dir
                / fold
                / model_class_to_long_lowercase_name[model_class]
                / "eval_results_test"
            )
            target_dir.parent.mkdir(parents=True, exist_ok=True)

            shutil.copytree(tmp_path / f"eval_results_{testcase}", target_dir)

            if fold == "validation":
                [results_path] = os.listdir(target_dir)

                with open(target_dir / results_path, "r") as f:
                    results = json.load(f)

                results["eval_args"]["fold"] = "VALIDATION"

                with open(target_dir / results_path, "w") as f:
                    json.dump(results, f)

        shutil.copytree(
            tmp_path / f"checkpoint_{testcase}",
            ensembling_dir / "input_checkpoints" / model_class_to_long_lowercase_name[model_class],
        )

    ensemble_checkpoint_dir = ensembling_dir / "output_checkpoints"

    run_with_python(
        path="./retrochimera/cli/optimize_ensembles.py",
        args=[
            f"data_dir={processed_output_dir}",
            f"results_dir={ensembling_dir}",
            f"output_dir={ensembling_dir}",
            f"input_checkpoints_dir={ensembling_dir / 'input_checkpoints'}",
            f"output_checkpoints_dir={ensemble_checkpoint_dir}",
            "dataset=test",
            "num_repeats=2",
        ],
    )

    with open(ensembling_dir / "individual.json", "rt") as f:
        individual_results = json.load(f)
        assert individual_results.keys() == set(
            model_class_to_long_lowercase_name[model_class] for model_class in model_classes
        )

    with open(ensembling_dir / "ensembles_optimized.json", "rt") as f:
        ensembling_results = json.load(f)
        assert len(ensembling_results) == 1

        chimera_result = ensembling_results["smiles_transformer_v1 template_localization_v0"]
        assert len(chimera_result) == 3

        # Check that ensembling weights look as expected.
        assert len(chimera_result["weights"]) == 2
        assert len(chimera_result["weights"][0]) == 100
        assert len(chimera_result["weights"][1]) == 100

        # Check that the ensemble achieves good accuracy.
        for fold in ["validation", "test"]:
            assert chimera_result[f"{fold}_accuracies"][0] > 0.85

    # Check that ensemble checkpoint was saved correctly.
    [ensemble_ckpt_dir] = os.listdir(ensemble_checkpoint_dir)
    assert ensemble_ckpt_dir == "ensemble_smiles_transformer_v1_template_localization_v0"

    ensemble_checkpoint_full_dir = ensemble_checkpoint_dir / ensemble_ckpt_dir
    assert get_num_files(ensemble_checkpoint_full_dir) == 3
    assert (ensemble_checkpoint_full_dir / "template_localization").is_dir()
    assert (ensemble_checkpoint_full_dir / "smiles_transformer").is_dir()

    with open(ensemble_checkpoint_full_dir / "models.json", "rt") as f:
        config = json.load(f)

        assert len(config) == 2
        for model_class in model_classes:
            (model_class_saved, weights) = config[convert_camel_to_snake(model_class)]

            assert model_class_saved == f"{model_class}Model"
            assert len(weights) == 100

    check_model_eval(
        eval_model_class="RetroChimera",
        model_dir=ensemble_checkpoint_full_dir,
        data_dir=processed_output_dir,
        eval_results_dir=tmp_path / "eval_results_ensemble",
    )



import torch
from torch import nn


class NormWrapper(nn.Module):
    """Wrapper fixing the behaviour of some of the normalization layers in PyTorch Geometric.

    The GNN models in PyTorch Geometric process batched graphs as if it was a single large graph,
    without knowing the batch assignment. However, some normalization layers (e.g. `GraphNorm`) need
    batch assignment as they normalize *within* each input graph. If such a layer is plugged into a
    model from `torch_geometric.nn.models`, it is silently used incorrectly, normalizing *across*
    graphs, which can have terrible consequences at inference time when batch size is different from
    training (see https://github.com/pyg-team/pytorch_geometric/discussions/7856 for more details).
    This class hacks PyG-s design and provides the batch assignment "through the back door".
    """

    # Classes in `torch_geometric.nn.norm` that need to be wrapped.
    NORMS_REQUIRING_BATCH_ASSIGNMENT = [
        "GraphNorm",
        "GraphSizeNorm",
        "InstanceNorm",
        "LayerNorm",
        "MeanSubtractionNorm",
        "PairNorm",
    ]

    def __init__(self, norm: nn.Module) -> None:
        super().__init__()

        self.norm = norm
        self.batch = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert self.batch is not None
        return self.norm(x, batch=self.batch)



import math
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Optional, Union

import torch
from more_itertools import batched
from syntheseus.reaction_prediction.utils.misc import parallelize
from torch import nn
from torch.nn import functional as f
from torch_geometric.data import Batch as BatchType

from retrochimera.chem.rules import RuleBase
from retrochimera.data.preprocessing.template_localization import (
    ProcessedSample,
    preprocess_samples,
)
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.encoders.base import get_encoder_by_name
from retrochimera.encoders.featurizers import RawRewriteGraph
from retrochimera.models.lightning import AbstractModel, log_classification_metrics
from retrochimera.utils.pytorch import split_select


@dataclass
class Batch:
    inputs: BatchType
    rewrites: BatchType
    targets: torch.Tensor
    targets_in_batch: torch.Tensor
    loc_targets: list[torch.Tensor]


class TemplateLocalizationModel(AbstractModel[TemplateReactionSample, ProcessedSample, Batch]):
    def __init__(
        self,
        input_encoder_class: str,
        rewrite_encoder_class: str,
        input_encoder_kwargs: dict[str, Any],
        rewrite_encoder_kwargs: dict[str, Any],
        classification_label_smoothing: float,
        localization_label_smoothing: float,
        classification_space_dim: Optional[int],
        free_rewrite_embedding_dim: int,
        classification_loss_type: str,
        classification_min_temperature: Optional[float],
        classification_max_temperature: Optional[float],
        negative_to_positive_targets_ratio: float,
        num_negative_rewrites_in_localization: int,
        n_classes: int,
        num_total_rewrite_lhs_atoms: int,
        rewrite_encoder_num_epochs: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert classification_loss_type in ["softmax", "sigmoid"]

        self.input_encoder = get_encoder_by_name(input_encoder_class)(**input_encoder_kwargs)
        self.rewrite_encoder = get_encoder_by_name(rewrite_encoder_class)(**rewrite_encoder_kwargs)

        if (classification_min_temperature is None) != (classification_max_temperature is None):
            raise ValueError(
                "Either both or neither of classification_{min,max}_temperature should be set"
            )

        self.classification_loss_type = classification_loss_type
        self.classification_min_temperature = classification_min_temperature
        self.classification_max_temperature = classification_max_temperature
        self.negative_to_positive_targets_ratio = negative_to_positive_targets_ratio
        self.num_negative_rewrites_in_localization = num_negative_rewrites_in_localization
        self.rewrite_encoder_num_epochs = rewrite_encoder_num_epochs

        if classification_loss_type == "softmax":
            self.classification_raw_temperature = None
        elif classification_loss_type == "sigmoid":
            # Initialize following "Sigmoid Loss for Language Image Pre-Training", unless a
            # temperature range is set, in which case initialize to the middle of that range.
            if classification_min_temperature is None:
                init_raw_temperature = math.log(10)
            else:
                init_raw_temperature = 0.0

            self.classification_raw_temperature = nn.Parameter(
                torch.full((1,), init_raw_temperature)
            )
            self.classification_bias = nn.Parameter(torch.full((1,), -10.0))
        else:
            raise ValueError(f"Unrecognized loss type {classification_loss_type}")

        assert self.input_encoder.mol_out_channels is not None
        assert self.rewrite_encoder.mol_out_channels is not None

        if free_rewrite_embedding_dim > 0:
            self.free_rewrite_embeddings = nn.Parameter(
                torch.zeros(n_classes, free_rewrite_embedding_dim)
            )
            rewrite_embedding_dim_total = (
                self.rewrite_encoder.mol_out_channels + free_rewrite_embedding_dim
            )
        else:
            self.free_rewrite_embeddings = None
            rewrite_embedding_dim_total = self.rewrite_encoder.mol_out_channels

        if classification_space_dim is not None:
            self.input_proj = nn.Linear(
                self.input_encoder.mol_out_channels, classification_space_dim
            )
            self.rewrite_proj = nn.Linear(rewrite_embedding_dim_total, classification_space_dim)
        else:
            assert self.input_encoder.mol_out_channels == rewrite_embedding_dim_total

            self.input_proj = nn.Identity()
            self.rewrite_proj = nn.Identity()

        if num_negative_rewrites_in_localization > 0:
            self.no_localization_embedding = nn.Parameter(
                torch.zeros(1, self.input_encoder.atom_out_channels)
            )
        else:
            self.no_localization_embedding = None

        self._classification_label_smoothing = classification_label_smoothing
        self._localization_label_smoothing = localization_label_smoothing
        self.n_classes = n_classes
        self.rulebase: Optional[RuleBase] = None
        self.rulebase_dir: Optional[Union[str, Path]] = None

        self.all_rewrites_batch_idx = nn.Parameter(
            torch.zeros(num_total_rewrite_lhs_atoms), requires_grad=False
        )
        self.all_rewrites_mol_outputs = nn.Parameter(
            torch.zeros(n_classes, self.rewrite_encoder.mol_out_channels), requires_grad=False
        )
        self.all_rewrites_atom_outputs = nn.Parameter(
            torch.zeros(num_total_rewrite_lhs_atoms, self.rewrite_encoder.atom_out_channels),
            requires_grad=False,
        )
        self._rewrite_encodings_up_to_date = False

        # Delay computing the rewrite graphs until we know we actually need it; we don't need it for
        # inference as we would have cached the results when saving the checkpoint.
        self._all_rewrite_graphs: Optional[list[RawRewriteGraph]] = None

    def set_rulebase(self, rulebase: RuleBase, rulebase_dir: Union[str, Path]) -> None:
        self.rulebase = rulebase
        self.rulebase_dir = rulebase_dir

    def _prepare_rule_weights(self) -> None:
        if self.classification_loss_type == "sigmoid":
            assert self.rulebase is not None

            positive_rule_frequency = torch.as_tensor(
                [rule.n_support for rule in self.rulebase.rules.values()], dtype=torch.float32
            )
            negative_rule_frequency = torch.ones_like(positive_rule_frequency)

            positive_rule_frequency /= positive_rule_frequency.sum()
            negative_rule_frequency /= negative_rule_frequency.sum()
            combined_rule_frequency = (
                positive_rule_frequency
                + self.negative_to_positive_targets_ratio * negative_rule_frequency
            )

            # Compute rule weights which are inversely proportional to frequency and average to 1.
            inverse_rule_frequency = 1 / combined_rule_frequency
            self._rule_weight = (
                inverse_rule_frequency / inverse_rule_frequency.sum() * self.n_classes
            )

    def _prepare_rewrite_graphs(self) -> None:
        if self._all_rewrite_graphs is None:
            assert self.rulebase is not None

            all_rewrites = [rule.rxn for rule in self.rulebase.rules.values()]
            self._all_rewrite_graphs = list(
                parallelize(self.rewrite_encoder.preprocess, all_rewrites, num_processes=0)
            )

    def on_fit_start(self) -> None:
        assert self.rulebase is not None

        self._prepare_rule_weights()
        self._prepare_rewrite_graphs()

        del self.rulebase

    def get_temperature(self) -> torch.Tensor:
        if self.classification_min_temperature is None:
            return torch.exp(self.classification_raw_temperature)
        else:
            assert self.classification_max_temperature is not None

            return self.classification_min_temperature + (
                self.classification_max_temperature - self.classification_min_temperature
            ) * torch.sigmoid(self.classification_raw_temperature)

    def forward_classification(
        self,
        input_reprs: torch.Tensor,
        rewrite_reprs: torch.Tensor,
        rewrite_ids: torch.Tensor,
        temperature: Optional[float] = None,
    ) -> torch.Tensor:
        if self.free_rewrite_embeddings is not None:
            rewrite_reprs = torch.cat(
                [rewrite_reprs, self.free_rewrite_embeddings[rewrite_ids]], dim=-1
            )

        input_reprs = self.input_proj(input_reprs)
        rewrite_reprs = self.rewrite_proj(rewrite_reprs)

        if self.classification_loss_type == "sigmoid":
            input_reprs = f.normalize(input_reprs)
            rewrite_reprs = f.normalize(rewrite_reprs)

        logits = torch.mm(input_reprs, torch.t(rewrite_reprs))

        if self.classification_loss_type == "sigmoid":
            if temperature is None:
                temperature = self.get_temperature()

            logits *= temperature
            logits += self.classification_bias

        return logits

    def forward_localization(
        self,
        input_graphs: BatchType,
        rewrite_graphs_node_in_lhs: torch.Tensor,
        rewrite_graphs_batch: torch.Tensor,
        input_graphs_atom_outputs: torch.Tensor,
        rewrite_graphs_atom_outputs: torch.Tensor,
        targets_in_batch: torch.Tensor,
        negative_targets_in_batch: Optional[torch.Tensor],
    ) -> list[torch.Tensor]:
        # Extract the representations of the lhs atoms, as these are the ones we need to localize.
        node_in_lhs_mask = rewrite_graphs_node_in_lhs.bool()
        rewrite_graphs_lhs_batch = torch.masked_select(rewrite_graphs_batch, node_in_lhs_mask)
        rewrite_graphs_lhs_atom_outputs = torch.masked_select(
            rewrite_graphs_atom_outputs, node_in_lhs_mask.unsqueeze(-1)
        ).view(-1, rewrite_graphs_atom_outputs.shape[-1])

        # Unbatch the atom representations to prepare for the localization step.
        input_atom_reprs_per_graph = split_select(input_graphs_atom_outputs, input_graphs.batch)
        rewrite_lhs_atom_reprs_per_graph = split_select(
            rewrite_graphs_lhs_atom_outputs, rewrite_graphs_lhs_batch
        )

        dot_prods_per_graph = []
        for graph_idx, graph_input_atom_reprs in enumerate(input_atom_reprs_per_graph):
            targets = [targets_in_batch[graph_idx]]

            if negative_targets_in_batch is not None:
                assert self.no_localization_embedding is not None
                graph_input_atom_reprs_ext = torch.cat(
                    [graph_input_atom_reprs, self.no_localization_embedding], dim=0
                )
                targets.extend(list(negative_targets_in_batch[graph_idx]))
            else:
                graph_input_atom_reprs_ext = graph_input_atom_reprs

            graph_rewrite_atom_reprs = torch.cat(
                [rewrite_lhs_atom_reprs_per_graph[idx] for idx in targets], dim=0
            )

            dot_prods = torch.mm(
                graph_rewrite_atom_reprs, torch.t(graph_input_atom_reprs_ext)
            )  # [num_loc_targets, num_atoms]
            dot_prods_per_graph.append(dot_prods)

        return dot_prods_per_graph

    def ttv_step(self, batch: Batch, step_name: str):
        input_graphs_enc = self.input_encoder(batch.inputs)

        batch_size = input_graphs_enc.mol_outputs.shape[0]
        device = input_graphs_enc.mol_outputs.device

        if step_name != "train" or (
            self.classification_loss_type == "softmax"
            and self.rewrite_encoder_num_epochs is not None
            and self.current_epoch >= self.rewrite_encoder_num_epochs
        ):
            if not self._rewrite_encodings_up_to_date:
                self._update_rewrite_encodings()

            # Use cached rewrite encodings for the entire rulebase.
            rewrite_graphs_mol_outputs = self.all_rewrites_mol_outputs
            rewrite_graphs_atom_outputs = self.all_rewrites_atom_outputs
            rewrite_graphs_node_in_lhs = torch.ones(
                self.all_rewrites_batch_idx.shape[0], device=device
            )
            rewrite_graphs_batch = self.all_rewrites_batch_idx

            # Include all templates in the loss.
            targets = torch.arange(self.n_classes, device=device)
            targets_in_batch = batch.targets[batch.targets_in_batch]

            # Call metrics "classification_loss", "acc", etc.
            metric_prefix = step_name
        else:
            self._rewrite_encodings_up_to_date = False

            # Call the rewrite encoder afresh as it is currently being trained.
            rewrite_graphs_enc = self.rewrite_encoder(batch.rewrites)
            rewrite_graphs_mol_outputs = rewrite_graphs_enc.mol_outputs
            rewrite_graphs_atom_outputs = rewrite_graphs_enc.atom_outputs
            rewrite_graphs_node_in_lhs = batch.rewrites.node_in_lhs
            rewrite_graphs_batch = batch.rewrites.batch

            # Include only the templates in the batch in the loss.
            targets = batch.targets
            targets_in_batch = batch.targets_in_batch

            # Call metrics "batch_classification_loss", "batch_acc", etc.
            metric_prefix = f"{step_name}_batch"

        classification_logits = self.forward_classification(
            input_reprs=input_graphs_enc.mol_outputs,
            rewrite_reprs=rewrite_graphs_mol_outputs,
            rewrite_ids=targets,
        )

        # `classification_targets` contains ones for correct input-rewrite combinations and zeros
        # elsewhere. Note that some columns (rewrites) may have multiple ones.
        targets_in_batch_flat = (
            targets_in_batch
            + torch.arange(batch_size, device=device) * classification_logits.shape[-1]
        )
        classification_targets = torch.zeros_like(classification_logits).ravel()
        classification_targets[targets_in_batch_flat] = 1.0
        classification_targets = classification_targets.reshape(classification_logits.shape)

        if self.classification_loss_type == "softmax":
            # Assymetric loss based on `https://arxiv.org/abs/2103.00020` (a symmetric loss would
            # not make sense as some rewrites may be negative samples with no matching inputs).
            classification_loss = f.cross_entropy(
                classification_logits,
                classification_targets,
                label_smoothing=self._classification_label_smoothing,
            )
        elif self.classification_loss_type == "sigmoid":
            # Loss variant based on `https://arxiv.org/abs/2303.15343`.
            # First compute appropriate weights for the pairwise losses.
            rule_weight = self._rule_weight.to(device)[targets]
            loss_weight = classification_targets + (
                1 - classification_targets
            ) * rule_weight.expand_as(classification_targets)

            # Convert {0, 1} to {-1, 1}.
            classification_targets = 2 * classification_targets - torch.ones_like(
                classification_targets
            )
            classification_pairwise_losses = -f.logsigmoid(
                classification_logits * classification_targets
            )

            if self._classification_label_smoothing > 0.0:
                # Label smoothing does not transfer directly to the sigmoid loss, but we do
                # something roughly similar: we make the "target" loss value a small positive
                # constant instead of 0.
                target_loss_value = -0.1 * math.log(1.0 - self._classification_label_smoothing)
                classification_pairwise_losses = torch.abs(
                    classification_pairwise_losses - target_loss_value
                )

            classification_loss = (loss_weight * classification_pairwise_losses).sum() / batch_size
        else:
            assert False

        assert input_graphs_enc.atom_outputs is not None
        assert input_graphs_enc.mol_outputs is not None

        if self.num_negative_rewrites_in_localization > 0:
            # Mask out the logits for the correct targets.
            classification_logits_masked = torch.clone(classification_logits.ravel())
            classification_logits_masked[targets_in_batch_flat] = -1e9
            classification_logits_masked = classification_logits_masked.reshape(
                classification_targets.shape
            )

            # Extract top *incorrect* targets for each input.
            negative_targets_in_batch = torch.topk(
                classification_logits_masked, k=self.num_negative_rewrites_in_localization
            ).indices
        else:
            negative_targets_in_batch = None

        # Run template localization for the ground-truth templates.
        dot_prods_per_graph = self.forward_localization(
            input_graphs=batch.inputs,
            rewrite_graphs_node_in_lhs=rewrite_graphs_node_in_lhs,
            rewrite_graphs_batch=rewrite_graphs_batch,
            input_graphs_atom_outputs=input_graphs_enc.atom_outputs,
            rewrite_graphs_atom_outputs=rewrite_graphs_atom_outputs,
            targets_in_batch=targets_in_batch,
            negative_targets_in_batch=negative_targets_in_batch,
        )

        localization_loss = torch.as_tensor(0.0, device=device)
        for graph_dot_prods, graph_loc_targets in zip(dot_prods_per_graph, batch.loc_targets):
            if self.num_negative_rewrites_in_localization > 0:
                graph_loc_targets_ext = torch.cat(
                    [graph_loc_targets, torch.zeros(len(graph_loc_targets), 1, device=device)],
                    dim=-1,
                )
                graph_loc_targets_negative = torch.zeros(
                    len(graph_dot_prods) - len(graph_loc_targets),
                    graph_loc_targets_ext.shape[-1],
                    device=device,
                )
                graph_loc_targets_negative[:, -1] = 1.0
                graph_loc_targets_ext = torch.cat(
                    [graph_loc_targets_ext, graph_loc_targets_negative], dim=0
                )
            else:
                graph_loc_targets_ext = graph_loc_targets

            graph_loc_targets_uniform = (
                torch.ones_like(graph_loc_targets_ext) / graph_loc_targets_ext.shape[-1]
            )
            graph_loc_targets_smooth = (
                (1.0 - self._localization_label_smoothing) * graph_loc_targets_ext
                + self._localization_label_smoothing * graph_loc_targets_uniform
            )
            localization_loss += f.kl_div(
                f.log_softmax(graph_dot_prods, dim=-1), graph_loc_targets_smooth, reduction="sum"
            )

        localization_loss /= batch_size * (self.num_negative_rewrites_in_localization + 1)
        loss = classification_loss + localization_loss

        if step_name == "train":
            if self.classification_loss_type == "sigmoid":
                self.log(
                    "classification_temperature", self.get_temperature(), batch_size=batch_size
                )
                self.log("classification_bias", self.classification_bias, batch_size=batch_size)

        self.log(f"{metric_prefix}_classification_loss", classification_loss, batch_size=batch_size)
        self.log(f"{metric_prefix}_localization_loss", localization_loss, batch_size=batch_size)
        self.log(f"{metric_prefix}_loss", loss, batch_size=batch_size)
        log_classification_metrics(
            log_fn=self.log,
            step_name=metric_prefix,
            batch_preds=classification_logits,
            batch_targets=targets_in_batch,
            batch_size=batch_size,
        )

        return loss

    def preprocess(
        self, samples: Iterable[TemplateReactionSample], num_processes: int = 0
    ) -> Iterable[ProcessedSample]:
        assert self.rulebase_dir is not None
        yield from preprocess_samples(
            samples=samples,
            rulebase_dir=self.rulebase_dir,
            input_encoder=self.input_encoder,
            num_processes=num_processes,
        )

    def collate(self, samples: list[ProcessedSample]) -> Batch:
        assert self._all_rewrite_graphs is not None

        targets = torch.as_tensor([sample.target for sample in samples], dtype=torch.long)
        targets_unique, targets_pos = torch.unique(targets, return_inverse=True)

        num_negative_targets = min(
            int(self.negative_to_positive_targets_ratio * len(targets_unique)),
            self.n_classes - len(targets_unique),
        )
        if num_negative_targets > 0:
            negative_target_weight = torch.ones(self.n_classes)
            negative_target_weight[targets_unique] = 0.0

            targets_unique = torch.cat(
                [
                    targets_unique,
                    torch.multinomial(negative_target_weight, num_samples=num_negative_targets),
                ]
            )

        return Batch(
            self.input_encoder.collate([sample.input for sample in samples]),
            self.rewrite_encoder.collate(
                [self._all_rewrite_graphs[target] for target in targets_unique]
            ),
            targets_unique,
            targets_pos,
            [torch.as_tensor(sample.loc_target, dtype=torch.float32) for sample in samples],
        )

    def _update_rewrite_encodings(self) -> None:
        assert self._all_rewrite_graphs is not None
        assert not self._rewrite_encodings_up_to_date
        self._rewrite_encodings_up_to_date = True

        # To get the best possible rewrite embeddings, we need to make sure we compute them in eval
        # mode, so that e.g. dropout is turned off. We restore the mode afterwards for transparency.
        orig_mode = self.training

        if orig_mode:
            self.eval()

        batch_idx_list: list[torch.Tensor] = []
        mol_outputs_list: list[torch.Tensor] = []
        atom_outputs_list: list[torch.Tensor] = []

        batch_idx_offset = 0
        device = next(self.parameters()).device

        REWRITES_ENCODING_BATCH_SIZE = 4096
        for graphs in batched(self._all_rewrite_graphs, REWRITES_ENCODING_BATCH_SIZE):
            batch = self.rewrite_encoder.collate(graphs)
            batch = self.rewrite_encoder.move_batch_to_device(batch, device=device)

            with torch.no_grad():
                batch_encoding = self.rewrite_encoder(batch)

            batch_idx_list.append(batch.batch.detach() + batch_idx_offset)
            mol_outputs_list.append(batch_encoding.mol_outputs.detach())
            atom_outputs_list.append(batch_encoding.atom_outputs.detach())

            batch_idx_offset += REWRITES_ENCODING_BATCH_SIZE

            del graphs, batch, batch_encoding

        if orig_mode:
            self.train()

        all_batch_idx = torch.cat(batch_idx_list)
        all_atom_outputs = torch.cat(atom_outputs_list)

        all_node_in_lhs = torch.cat(
            [torch.as_tensor(graph.node_in_lhs) for graph in self._all_rewrite_graphs]
        )
        all_node_in_lhs = all_node_in_lhs.bool().to(device)

        self.all_rewrites_batch_idx.data = torch.masked_select(all_batch_idx, all_node_in_lhs)
        self.all_rewrites_mol_outputs.data = torch.cat(mol_outputs_list)
        self.all_rewrites_atom_outputs.data = torch.masked_select(
            all_atom_outputs, all_node_in_lhs.unsqueeze(-1)
        ).view(-1, all_atom_outputs.shape[-1])

    def on_save_checkpoint(self, checkpoint: dict[str, Any]) -> None:
        if not self._rewrite_encodings_up_to_date:
            self._update_rewrite_encodings()

            # If we updated encodings *after* `checkpoint` was assembled then refresh them here.
            for key in [
                "all_rewrites_batch_idx",
                "all_rewrites_mol_outputs",
                "all_rewrites_atom_outputs",
            ]:
                checkpoint[key] = getattr(self, key).detach().cpu()



import pytest


@pytest.fixture(scope="module")
def tmp_path(tmp_path_factory):
    """Fixture to create a temporary directory shared across tests."""
    yield tmp_path_factory.mktemp("shared")


@pytest.fixture(scope="module")
def deduplicate_output_path(tmp_path):
    return tmp_path / "output.smi"


@pytest.fixture(scope="module")
def split_output_dir(tmp_path):
    split_dir = tmp_path / "split"
    split_dir.mkdir()
    return split_dir


@pytest.fixture(scope="module")
def processed_output_dir(tmp_path):
    return tmp_path / "processed"


@pytest.fixture(scope="module")
def augmented_output_dir(tmp_path):
    return tmp_path / "augmented"
